// C++ Dependencies
using namespace std;
# include <iostream>
# include "GocrSegment.h"
# include "GocrModule.h"
# include "Page.h"
# include "Glyph.h"
# include "Line.h"
# include "Image.h"

// C Dependencies
# include "gocr.h"     // Job
# include "otsu.h"     // otsu, thresholding
# include "barcode.h"  // detect_barcode
# include "pgm2asc.h" // pgm2asc, getTextLine

extern int progress(int p1, int p2);
extern int scan_boxes( pix *p );
extern int add_line_info();
extern int sort_box_func (const void *a, const void *b);
extern int glue_broken_chars( pix *pp );
extern void measure_pitch( Job *job );
extern int find_same_chars( pix *pp);
extern int char_recognition( pix *pp, int mo);
extern int compare_unknown_with_known_chars(pix * pp, int mo);
extern int  try_to_divide_boxes( pix *pp, int mo);
extern int list_insert_spaces( pix *pp, Job *job );
extern int context_correction( Job *job );
extern int  divide_vert_glued_boxes( pix *pp, int mo);
// FUTURE FIX: Most of the above are NOT needed in this class!!

namespace Conjecture {
  GocrSegment::GocrSegment() {
    // Default Constructor
  }

  bool GocrSegment::execute(Page* page) {
    //DEBUG: cerr << "**** HERE in GocrSegment::partition" << endl;

    // Obtain the env
    const Env* env = page->env();

    // Obtain the module
    const GocrModule* module = dynamic_cast<const GocrModule*>(env->module());
    if ( module == NULL ) {
	throw string("IncompatibleModuleException: GocrFormat requires the GocrModule module");
    }

    // Obtain the ubiquitous Job structure
    Job* job = module->job();

    // Initialize its src.p from the image.
    const Image* image = page->image();
    job->src.p = *(image->data());

    // Set up some local vars
    int  verbose = job->cfg.verbose;
    int  mode    = job->cfg.mode;
    pix* pp      = &(job->src.p);
    bool result  = true;

    // ******************************************************
    // Count Colors (create histogram)
    //  - used to create a upper and lower limit for cs
    //  - cs is the optimum gray value between cs_min and cs_max
    //  - also inverse scans could be detected here later
    if (job->cfg.cs==0) {
      job->cfg.cs=otsu( pp->p,pp->y,pp->x,0,0,pp->x,pp->y, job->cfg.verbose & 1 );
    }
    // renormalize the image and set the normalized threshold value
    job->cfg.cs=thresholding( pp->p,pp->y,pp->x,0,0,pp->x,pp->y, job->cfg.cs );
    if ( verbose ) {
      cerr << "# thresholding new_threshold= %d" << job->cfg.cs << endl;
    }
    progress(5,0); /* progress is only estimated */
    job->tmp.ppo=job->src.p; /* temporarely, removed later */
  
    // ******************************************************
    // Load character database
    if ( mode & 2 ) { load_db(); }
    progress(8,0); /* progress is only estimated */

    // ******************************************************
    // Frame letters
    //  - this is first step for reorganize the PG
    //  - look for letters, put rectangular frames around letters
    //  - letter = connected points near color F
    //  - should be used by dust removing (faster) and line detection!
    //  - 0..cs = black letters, last change = Mai99
    scan_boxes( pp );
    if ( !job->res.numC ) { 
      cerr << "# no boxes found - stopped" << endl;
      if ( verbose & 32 ) { debug_img("out01",job,8); }
      result = false;
      goto EndProcess;
    }
    progress(10,0); /* progress is only estimated */

    // ******************************************************
    // Clean up noise in the image
    //  - iterates over the boxlist.
    remove_dust( job );

    // ******************************************************
    // Border smoothing of characters >7x16
    smooth_borders( job );
    progress(12,0); /* progress is only estimated */

    // ******************************************************
    // Detect barcode
    //  - This would seem to be a rather specialized application?
    //    Most images don't have barcodes in them, why are we always
    //    checking?
    detect_barcode( job );  /* mark barcode */

    // ******************************************************
    // Establishes which boxes in boxlist are pictures.
    //  - for example, any box whose size is 4 times larger in
    //    either direction than the average box size.
    //  - the box is marked as PICTURE if it is a picture.
    detect_pictures( job );

    // ******************************************************
    // Remove boxes marked as pictures from further consideration
    //  - done early so we don't waste time processing them.
    remove_pictures( job );

    // ******************************************************
    // Establish whether the image is already vertically aligned,
    // and if not, what its rotation angle is.
    detect_rotation_angle( job );

    // ******************************************************
    // Establish horizontal lines of characters
    //  - marks up the job->tmp.ppo image for debugging
    detect_text_lines( pp, job->cfg.mode ); 
    progress(20,0); /* progress is only estimated */
    
    // ******************************************************
    add_line_info(/* &(job->res.boxlist) */);
    if(job->cfg.verbose&32) debug_img("out10",job,4+8);
    
    // ******************************************************
    // Detect and divide vertically stacked boxes that touch
    // each other, but which have been established to belong
    // to different lines.
    //  - this must be performed after add_line_info, 
    //    before list_sort
    divide_vert_glued_boxes( pp, job->cfg.mode);
    
    // ******************************************************
    // Characters can be joined together due to serifs.  This
    // attempts to detect such situations and split the box
    // into two (or more).
    //   - makes corrections on pixmap 
    remove_melted_serifs( pp );
    
    // ******************************************************
    // Sort the box list, left to right, top to bottom.
    list_sort(&(job->res.boxlist), sort_box_func);
    
    // ******************************************************
    // Merges boxes that are part of the same character
    //  - a box completely contained within another
    //    box is merged into the surrounding box
    glue_broken_chars( pp );

    // ******************************************************
    //  - removes black borders around bad scans.
    remove_rest_of_dust( );
    
    // ******************************************************
    // Measure pitch
    //  - detect monospaced fonts
    //  - measure overall and per-line pitch
    measure_pitch( job );

    // ******************************************************
    if (mode & 64) { find_same_chars( pp ); }
    progress(30,0); /* progress is only estimated */

    EndProcess:
    return(result);

  }
};

