# ifndef Conjecture_PixelLine_h
# define Conjecture_PixelLine_h Conjecture_PixelLine_h

// Insert SVN header substitution variable here
// Created by wade
// Class Champion: Unassigned

# include "Root.h"  // parent

namespace Conjecture {
    class OptIm;  // friend declaration
    
    /** ****************************************************************
     * \class PixelLine
     *
     * \brief A class for describing a line of up to 32 pixels.
     *
     * This class is used within OptIm to provide efficient feature
     * extraction.
     *
     ******************************************************************/

    // IMPLEMENTATION NOTES:
    //   - this class is designed for highly efficient space utiliation,
    //     while also providing support for highly efficient feature
    //     analysis.
    //   - some "auxillary" methods are quite slow (i.e. 'set' and
    //     'clear' - this class is meant to be used primarily on
    //     frozen images, so setting and clearing bits is not a
    //     priority for efficient implementation.
    //   - does NOT store the size of the pixel line (assumed to be
    //     be maintained elsewhere). This implies that the number of
    //     white pixels cannot be reported without obtaining the size
    //     (#black is stored, and #white = total - #black).
    //   - not using setters internally in some situations because
    //     bit-specifiers cause temporaries to be used in returns

    class PixelLine : public Root {
      public:
        static const u1 MaxPixels = 31;

        // Constructors/Destructors
        PixelLine(u4 pixels);
        
        // Accessors
        inline u1   holes()  const { return this->_holes;  }
        inline u4   pixels() const { return this->_pixels; }
        inline u1   black()  const { return this->_black;  }
        
        // Input/Output
        
        // Interface

        /**
         * Return 0 if pixel position i is not set, 1 if it is set.
         *
         * Pixels start indexing from 0 on the RIGHT! The 'index' is
         * is NOT verified to be within either soft or hard size
         * limits.
         *
         */
        inline bool get(unsigned i) const {
            return (this->pixels() & (1<<i)) ? true : false;
        }
        
        /**
         * Sets pixel i to 1.
         * 
         * Pixels start indexing from 0. The 'index' is NOT verified
         * to be within either soft or hard size limits. This method
         * is NOT efficient, as it requires the updating of
         * meta-information in addition to the pixel itself.
         *
         */
        void set(unsigned i);

        /**
         * Sets pixel i to 0.
         * 
         * Pixels start indexing from 0. The 'index' is NOT verified
         * to be within either soft or hard size limits. This method
         * is NOT efficient, as it requires the updating of
         * meta-information in addition to the pixel itself.
         *
         */
        void clear(unsigned i);
        
        /**
         * Binary representation, as a string (inefficienct -
         * debugging only).
         *
         */
        std::string str(char on = '#', char off = '.', u1 size = 31) const;
        inline std::string str(u1 size) const { return this->str('#','.',size); }

	/**
	 * Assign this PixelLine from another.
	 *
	 * Cannot use operator= because we want to support
	 * pixel shifting via 'rightdel'.
	 *
	 * \param rightdel Number of bits to shift the 'pixels' data
	 * to the right during the copy.
	 *
	 */
        void assign(const PixelLine& other, u1 rightdel = 0);

        static void test(int argc = 0, const char* argv[] = NULL);

      protected:
        // Accessors
        inline const bool &    updated() const { return this->_updated; }
        inline void            updatedIs(const bool & updated) { this->_updated = updated; }
        inline void            blackIs(const u1 & black) { this->_black = black; }
        inline void            holesIs(const u1 & holes)   { this->_holes = holes; }
        inline void            pixelsIs(const u4 & pixels) { this->_pixels = pixels; }
        
        // Methods 
        inline void rawset(unsigned i)   { this->_pixels |= (1<<i);  }
        inline void rawclear(unsigned i) { this->_pixels &= ~(1<<i); }

        inline void rawline(u4 pixels) { this->_pixels = pixels; this->updatedIs(false); }
        
      private: 
	/**
	 * PixFilt needs access to rawset and rawclear for efficiency
	 *
	 */
        friend class PixFilt;

	/**
	 * OptIm needs to set the pixel data
	 *
	 */
        friend class OptIm;

        // Accessors
        
        // Methods 
        void updateMeta();

        // State

	/**
	 * The sequence of pixels in this PixelLine.
	 *
	 * Each pixel is stored as a single bit, with 1 representing
	 * black and 0 representing white. A maximum of 31 pixels is
	 * allowed (see 'black' below). For PixelLines of size less
	 * than 31, pixel positions beyond the soft size are set to
	 * 0.
	 *
	 */
        u4              _pixels;

	/**
	 * The number of occurrences of a contiguous sequence of white
	 * pixels is surrounded by black pixels. 
	 * 
	 * White pixels on the boundary do not count.  Note that the
	 * maximum number of occurances occurs when the pixel line
	 * consists of alternating black and white pixels like this:
	 *   1010101010101010101010101010101
	 * and in this situation, there are 15 holes (15 places where
	 * one or more 0's is between 1's).  We can represent the
	 * value 15 in 4 bits.
	 *
	 */
        unsigned int    _holes:4;

	/**
	 * The number of black pixels (number of 1 bits).
	 *
	 * There are a maximum of 31 pixels, and thus a maximum
	 * of 31 black pixels.  We can represent 31 in 5 bits.
	 *
	 */
        u1              _black:5;

	/**
	 * True if this PixelLine's metadata is up-to-date, false otherwise
	 *
	 * A PixelLine's metadata is NOT made up-to-date on creation.
	 * Instead, 'updateMeta' must be invoked.  Certain bit-modifying
	 * methods may 
	 *
	 */
        bool            _updated;

	/**
	 * FUTURE FIX: There are 22 bits still available. If
	 * bit-packing works for 64-bit longlongs, one possibility
	 * would be to extend the maximum size. We would need to
	 * increase the size of 'holes' and 'black' by one each
	 * (they'd take up 11 bits). This allows us to store anywhere
	 * between 31 and 52 pixels in _pixels. We must establish how
	 * common glyphs larger than 31 pixels are!!
	 */
    };
}

# endif // Conjecture_PixelLine_h

