using namespace std;
# include <iostream>
# include "WmhSegment.h"
# include "GocrModule.h"

// C dependencies (temporary?)
# include "gocr.h"  // Job
# include "list.h"  // struct list = List from gocr
# include "pgm2asc.h" // pgm2asc, getTextLine

// Include
# include "Glyph.h"
# include "GocrSegment.h"
# include "Page.h"
# include "Line.h"

namespace Conjecture {

    WmhSegment::WmhSegment() { }

    // Interface
    bool WmhSegment::execute(Page* page) {
        bool result = false;
        
        static GocrSegment* gocr = NULL;
        if ( gocr == NULL ) {
            gocr = new GocrSegment();
        }
        result = gocr->execute(page);

        // Based on the above Conjecture results, we can fully initialize
        // the fields in our C++ classes.
        //  - job->res.boxlist is a collection of Box instances.
        //  - Each Glyph is wrapped around a Box (for now)
        //  - Each Glyph exists within a Word or Line or Region or Page
        //    (usually a Line, for now)
        //  - We may not support Word's to begin with, but when we do,
        //    they will usually be within Lines
        //  - Most documents will contain Lines, but some special-case
        //    "scatterplot" documents in which words are thrown randomly
        //    onto the page won't.
        //  - For now, it is probably easiest to pre-process images
        //    so that there is only one Region per Page (and thus no need
        //    for a Region).  However, advanced support should be able
        //    to detect two-column documents, embedded images, etc.,
        //    in which case each Page will consist of multiple Regions.
        //
        // For now, we iterate over each Box, create an associated Glyph,
        // and accumulate Glyphs into Line instances, and accumulate
        // Lines into the Page.  We do not currently use Word or Region.
        
        if ( result ) {
	    const Env* env = page->env();
	    const GocrModule* module = dynamic_cast<const GocrModule*>(env->module());
	    if ( module == NULL ) {
		throw string("IncompatibleModuleException: WmhFormat requires GocrModule");
	    }
	    Job* job = module->job();

            List* glist = &(job->res.boxlist);
            // 'start' is a dummy node - we advance to first meaningful one.
            ::Element* elem = glist->start.next;
            Line*    line = NULL;
            int tbaseline = 0;
            int n         = glist->n;
            int i         = 0;
            
            // Iterate over the list of boxes
            Glyph* glast = NULL;

            while ((elem) && (i<n)) {
                // Obtain the Box
                Box* box = (Box*)(elem->data);
                if ( box ) {
                    
                    // Create the Glyph around the box
                    Glyph* g       = new Glyph(page, box);
                    int baseline   = box->m3;
                    if ( baseline != tbaseline ) {
                        // Looks like a new line.
                        line         = new Line(page, Coord(0,0), Coord(0,0), box->m4, box->m3, box->m2, box->m1);
                        tbaseline    = baseline;
                        page->registerElement(line);
                    }

                    // NOTE: There are certain heuristics applied here
                    // to filter out erroneous glyphs.  There are two
                    // ways to deal with them:
                    //   1) Do not register them with the line
                    //   2) Register them, but mark them invisible.
                    // For now, it is easier to mark them invisible,
                    // but deletion may make Word identification, etc.
                    // easier - FUTURE FIX.
                    if ( ( glast != NULL ) && 
                         // ( g->topleft().withinBounds(glast->topleft(), glast->bottomright()) )
                         ( (g->topleft().x - glast->topleft().x) < 3 )
                        ) {
                        // The gocr segmentation sometimes returns
                        // glyphs for the same region. Not sure if
                        // this is a feature or a bug.  Although we
			// could process both and see which one returns
			// a more confident match, for now, we just keep
			// the larger of the two.
                        cerr << "NOTE: Found a glyph contained within the bounds of the previous glyph" << endl;
                        glast->printSummary(cerr,"",i-1);
                        g->printSummary(cerr,"",i);
			if ( g->size() < glast->size() ) {
			    cerr << "NOTE:   Dropping current glyph in favor of previous" << endl;
			    g->markInvisible();
			} else {
			    cerr << "NOTE:   Dropping previous glyph in favor of current" << endl;
			    glast->markInvisible();
			}
                    } else if ( ( g->size().x <= 1 ) && ( g->size().y <= 1 ) ) {
                        // gocr segmentation sometimes provides single pixel glyphs??
                        // FUTURE FIX: REMOVE THIS CODE when we make our own segmenter!
                        
                    }
                    // We always add the glyph (but it might be marked invisible)
                    line->registerElement(g);

                    // Remember the last glyph
                    glast = g;
                } else {
                    cerr << "Glyph " << i << " has no data" << endl;
                }
                elem = elem->next;
                ++i;
            }
            
            if ( page->env()->verbose() > 4 ) {
                cerr << "***** NOTE: Verbose is " << page->env()->verbose() << endl;
                page->printStructure(cerr);
            }
        }
        return result;
    }
};

