using namespace std;
# include <iostream>
# include <iomanip>
# include <sstream>     // stringstream
# include "PixelLine15.h" // myself

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

// Include

namespace Conjecture {

    PixelLine15::PixelLine15(unsigned int p) {
	// Only allow 15 pixels
	p &= 0x7fff;
	
	// Store the pixels.  Pixel 0 is right-most pixel.
	this->pixelsIs(p);
	
	this->updateMeta();
    }

    void PixelLine15::updateMeta() {
	u2 p = this->pixels();

	// Temp vars
	unsigned int black  = 0;
	unsigned int flips  = 0;
	bool         laston = false;
	
	// Iterate: check low-order bit, shift right one.
	while (p) {
	    if ( p&0x1 ) {
		// Pixel is black
		++black;
		if (!laston) { ++flips; }
		laston = true;
	    } else {
		// Pixel is white
		if (laston) { ++flips; }
		laston = false;
	    }
	    // shift right one
	    p >>= 1;
	}
	
	// The number of holes is half the number of flips:
	unsigned int holes = int(flips/2);
	this->blackIs(black);
	this->holesIs(holes);
    }
    
    // Interface
    string PixelLine15::str(char on, char off, unsigned char size) const {
	stringstream s;
	s << setw(2) << this->holes() << " ";
	for (int i = size-1; i >= 0; --i) {
	    if ( this->get(i) ) {
		s << on;
	    } else {
		s << off;
	    }
	}
	s << " " << left << setw(2) << this->black();
	return s.str();
    }
    
    void PixelLine15::set(unsigned i)  {
	// establish current value.
	bool on = this->get(i);
	
	if (!on) {
	    // We are switching from white to black
	    //  - increment black count
	    this->blackIs(this->black()+1);
	    //  - if we have '1X1' (X=0), reduce hole count by 1
	    //  - if we have '1X01' or '10X1', no affect on hole count
	    //  - if we have '10X01' or '00X01' or '10X00', increase hole count by 1
	    bool left  = (i<15) ? this->get(i+1) : false;
	    bool right = (i>0)  ? this->get(i-1) : false;
	    int adj = 0;
	    if ( left && right ) { adj = -1; }
	    else if ( (!left)&&(!right) ) { adj = 1; }
	    if ( adj ) { this->holesIs(this->holes()+adj); }
	    
	    // Make this pixel black
	    this->_pixels |= (1<<i);
	}
    }
    
    void PixelLine15::clear(unsigned i)  {
	// establish current value.
	bool on = this->get(i);
	
	if (on) {
	    // We are switching from black to white
	    //  - increment black count
	    this->blackIs(this->black()-1);
	    //  - if we have '1X1', increase hole count by 1
	    //  - if we have '0X1' or '1X0', no hole affect
	    //  - if we have '10X01 or '00X01' or '10X00', decrease hole count by 1
	    bool left  = (i<15) ? this->get(i+1) : false;
	    bool right = (i>0)  ? this->get(i-1) : false;
	    int adj = 0;
	    if ( left && right ) { adj = 1; }
	    else if ( (!left)&&(!right) ) { adj = -1; }
	    if ( adj ) { this->holesIs(this->holes()+adj); }
	    
	    // Make this pixel white
	    this->_pixels ^= (1<<i);
	}
    }
    
    void PixelLine15::test(int argc, const char* argv[]) {
	// 0100010010110011
	PixelLine15 p1(0x44b3);
	cout << p1.str() << endl;
	p1.set(2);
	cout << p1.str() << endl;
	p1.set(3);
	cout << p1.str() << endl;
	p1.set(3);
	cout << p1.str() << endl;
	p1.clear(3);
	cout << p1.str() << endl;
	p1.clear(3);
	cout << p1.str() << endl;
	p1.clear(2);
	cout << p1.str() << endl;
	
	PixelLine15 p2(0x15);
	cout << p2.str('*','.') << endl;
	p2.set(1);
	cout << p2.str('*','.') << endl;
	p2.set(3);
	cout << p2.str('*','.') << endl;
	p2.set(3);
	cout << p2.str('*','.') << endl;
	p2.clear(3);
	cout << p2.str('*','.') << endl;
	p2.clear(3);
	cout << p2.str('*','.') << endl;
	p2.clear(1);
	cout << p2.str('*','.') << endl;
	
	PixelLine15 p3(0x5);
	cout << p3.str('#',' ') << endl;
	p3.set(2);
	cout << p3.str('#',' ') << endl;
	p3.set(1);
	cout << p3.str('#',' ') << endl;
	p3.clear(1);
	cout << p3.str('#',' ') << endl;
	p3.clear(2);
	cout << p3.str('#',' ') << endl;
	
	PixelLine15 p4(0xa);
	cout << p4.str() << endl;
	
	PixelLine15 p5(0x80);
	cout << p5.str() << endl;
    }
};

