# ifndef PixelLine15_h
# define PixelLine15_h PixelLine15_h

// ####################################################################
// # Conjecture: An Extensible Optical Character Recognition Framework  #
// #                                                                  #
// # Copyright: Lesser GNU Public License (LGPL)                      #
// #                                                                  #
// # Overview:                                                        #
// #  - Conjecture is a collection of C++ classes, and an OCR engine    #
// #  - Conjecture is designed to allow customization at all levels     #
// #    and to encourage individuals to contribute incremental        #
// #    improvements in algorithms.                                   #
// #  - Overall design discussions can be found in                    #
// #      $SNROOT/docs/doxygen/html/index.html                        #
// #  - Conjecture is designed to interact with and build on other      #
// #    open-source OCR programs.                                     #
// ####################################################################

# include "Root.h"  // parent
# include "Coord.h"

namespace Conjecture {

  // ##################################################################
  /// \class PixelLine15
  ///
  /// \brief A class for describing a line of up to 16 pixels.
  ///
  /// This class is used within OptIm to provide efficient feature
  /// extraction.
  // ##################################################################

  // IMPLEMENTATION NOTES:
  //   - this class is designed for highly efficient space utiliation,
  //     while also providing support for highly efficient feature
  //     analysis.
  //   - some "auxillary" methods are quite slow (i.e. 'set' and
  //     'clear' - this class is meant to be used primarily on
  //     frozen images, so setting and clearing bits is not a
  //     priority for efficient implementation.
  //   - does NOT store the size of the pixel line (assumed to be
  //     be maintained elsewhere). This implies that the number of
  //     white pixels cannot be reported without obtaining the size
  //     (#black is stored, and #white = total - #black).
  //   - not using setters internally in some situations because
  //     bit-specifiers cause temporaries to be used in returns

  class PixelLine15 : public Root {
  public:
    // Constructors/Destructors
    PixelLine15(unsigned int pixels);

    // **************
    // Accessors
    inline unsigned int holes()  const { return this->_holes;  }
    inline unsigned int pixels() const { return this->_pixels; }
    inline unsigned int black()  const { return this->_black;  }
    
    // **************
    // Input/Output
    
    // **************
    // Interface

    /// Return 0 if pixel position i is not set, 1 if it is set.
    ///
    /// Pixels start indexing from 0. The 'index' is is NOT verified
    /// to be within either soft or hard size limits.
    inline bool get(unsigned i) const {
      return (this->pixels() & (1<<i)) ? true : false;
    }

    /// Sets pixel i to 1.
    /// 
    /// Pixels start indexing from 0.  The 'index' is NOT verified
    /// to be within either soft or hard size limits.
    void set(unsigned i);

    /// Sets pixel i to 0.
    /// 
    /// Pixels start indexing from 0.  The 'index' is NOT verified
    /// to be within either soft or hard size limits.
    void clear(unsigned i);

    /// Binary representation, as a string (inefficienct - debugging only).
    std::string str(char on = '#', char off = '.', unsigned char size = 15) const;

    static void test(int argc = 0, const char* argv[] = NULL);

  protected:
    // **************
    // Accessors
    inline void   holesIs(unsigned int holes)   { this->_holes  = holes;  }
    inline void   pixelsIs(unsigned int pixels) { this->_pixels = pixels; }
    inline void   blackIs(unsigned int black)   { this->_black  = black;  }
    
    // **************
    // Methods 
    
  private: 
    // **************
    // Accessors
    
    // **************
    // Methods 
    void updateMeta();

    // **************
    // State

    /// The sequence of pixels in this PixelLine15.
    ///
    /// Each pixel is stored as a single bit, with 1 representing
    /// black and 0 representing white. A maximum of 15 pixels is
    /// allowed (see 'black' below). For PixelLine15s of size less than
    /// 15, pixels beyond the size are 0.
    unsigned int    _pixels:15;

    /// True if this PixelLine15's metadata is up-to-date, false otherwise.
    ///
    /// The placement of this boolean field is important, as it provides
    /// better internal alignment.
    unsigned int    _clean:1;

    /// The number of black pixels (number of 1 bits).
    unsigned int    _black:4;

    /// The number of occurrences of a contiguous sequence of white
    /// pixels is surrounded by black pixels. 
    /// 
    /// White pixels on the boundary do not count.  Note that the
    /// maximum number of occurances occurs when the pixel line
    /// consists of alternating black and white pixels like this:
    ///   101010101010101
    /// and in this situation, there are 7 holes (7 places where
    /// one or more 0's is between 1's).  We can represent the
    /// value 7 in 3 bits.
    unsigned int    _holes:3;

    // FUTURE FIX: There are still 8 bits remaining that can be
    // filled with other information useful to feature analysis.
    // Possibilities include:
    //   - 2 bits to indicate whether the left-most black pixel
    //     is at left border (0), 1 in from left border (1),
    //     2 in from left border (2), or 3+ in from left border
    //     (3).  An analogous 2 bits for right-most black pixel
    //     as well.  Leaves 4 bits for other things.
    //   - index of left-most black pixel (4 bits) and index of
    //     right-most black pixel (4 bits).

  };
}

# endif // PixelLine15_h

