# ifndef Conjecture_Debug_h
# define Conjecture_Debug_h Conjecture_Debug_h

// Insert SVN header substitution variable here
// Class Champion: Unassigned

#include "Root.h"
#include <iostream>
#include <sstream>

namespace Conjecture {

    /**
     * Debug class.
     */
    class Debug : public Root {
        public:

            /**
             * Verbose levels.
             */
            typedef enum {
                DEBUG_LEVEL_NONE          = 0x00,   ///< No output
                DEBUG_LEVEL_ERROR         = 0x01,   ///< Print errors
                DEBUG_LEVEL_WARNING       = 0x02,   ///< Print warnings
                DEBUG_LEVEL_INFO          = 0x04,   ///< Print information
               	DEBUG_LEVEL_IMAGES_BASIC  = 0x40,   ///< Output basic images
                DEBUG_LEVEL_IMAGES_FULL   = 0x80,   ///< Output all images
                DEBUG_LEVEL_ALL           = 0xFF    ///< Output everything
            } verboseLevel;

            /**
             * Singleton access method.
             *
             * @return A reference to this singleton.
             */
            static Debug &getHandle(void) {
                static Debug *singleton = new Debug();
                return *singleton;
            }

            /**
             * Sets the target log stream.
             *
             * @param s The std::stream to use for output.
             */
            inline void setLogStream(std::ostream &s) {
                this->_outStream = &s;
            }

            /**
             * Toggles verbose flags.
             *
             * @param l The flags to toggle.
             */
            inline void toggleLevel(verboseLevel l) {
                this->_level = static_cast<verboseLevel>(l | this->_level);
            }

            /**
             * Untoggles verbose flags.
             *
             * @param l The flags to toggle.
             */
            inline void untoggleLevel(verboseLevel l) {
                this->_level = static_cast<verboseLevel>((~l) & this->_level);
            }

            /**
             * Sets verbose flags.
             *
             * @param l The flags to toggle.
             */
            inline void setLevel(verboseLevel l) {
                this->_level = l;
            }

            /**
             * Checks if a given level should be printed.
             *
             * @return A boolean value.
             */
            inline bool isInLevel(verboseLevel l) {
                if ((l & this->_level) != 0) {
                    return true;
                }
                return false;
            }

            /**
             * Prints a header with standard information.
             *
             * @param l The verbose level of this message.
             */
            virtual void header(verboseLevel l);

	    /**
	     * Overloads << operator.
	     *
	     * @param str The std::string to output.
	     * @return A reference to this class.
	     */
	    virtual Debug & operator<< (const std::string &str);
	    
	    /**
	     * Overloads << operator.
	     *
	     * @param str The character string to output, ended by \\0.
	     * @return A reference to this class.
	     */
	    virtual Debug & operator<< (const char *str);

        protected:
            verboseLevel            _level;           /**< Current flags */

            /**
             * Accessor for the outStream field.
             *
             */
            inline std::ostream* outStream() { return this->_outStream; }

	    /**
             * Converts a verboseLevel to a string.
	     \param l The level.
	     \return The name of that level.
	    */
	    inline const char *levelToString(verboseLevel l) const {
		switch (l) {
		  case DEBUG_LEVEL_ERROR:
		    return (const char *)"ERROR:: ";
		  case DEBUG_LEVEL_WARNING:
		    return (const char *)"WARNING: ";
                  case DEBUG_LEVEL_INFO:
                    return (const char *)"Info: ";
                  case DEBUG_LEVEL_IMAGES_BASIC:
                    return (const char *)"Image (basic): ";
                  case DEBUG_LEVEL_IMAGES_FULL:
                    return (const char *)"Image (full): ";
                  default:
                    return (const char *)"Unknown: ";
		}
		// should never happen
		return NULL;
	    }

            virtual ~Debug();

      private:
            std::ostream            *_outStream;      /**< Output stream */

            /**
             * Private constructor. This class is a singleton.
             */
            Debug();

            /**
             * Avoids assignment. This class is a singleton.
             */
            void operator=(Debug &);

            /**
             * Avoid cloning.  This class is a singleton.
             */
            Debug(const Debug &);
    };

#ifdef ENABLE_LOG
 #define LOG(level, message) \
    if (Debug::getHandle().isInLevel(Conjecture::Debug::level)) { \
            Debug::getHandle().header(libglass::Log::level); \
            std::stringstream ___s; \
            ___s << message; \
            l << ___s.str(); \
        } \
    }
#else
 #define LOG(l, level, message)
#endif


} // namespace

#endif
