
// C++ Dependencies
# include <iostream>
# include <iomanip>
# include <sstream>
# include <string>    
# include <vector>   
# include "Element.h"  
# include "Page.h"  // to define 'image()'
# include "Image.h"
# include "Glyph.h" // magnifyGlyphs

// C dependencies (temporary)
# include "pnm.h"

using namespace std;

namespace Conjecture {
    
    Element::Element(Element* parent, const Coord& topleft, const Coord& bottomright) : _topleft(topleft), _bottomright(bottomright) {
	// Deal with parent issues
	if ( parent != NULL ) {
	    // This doesn't work as desired because the vtbl hasn't been
	    // fully initialized yet. The goal is to have 'this->type()'
	    // invoke the subclass definition (and return an integer between
	    // 1 and 5), but because it is a virtual method being invoked
	    // within the constructor, the local definition is invoked
	    // instead.  May need to add a 'validate' method to do this
	    // checking.  As long as we don't pass in incorrect parent
	    // types this code isn't necessary, so it isn't critical.
	    //
	    //  FIXTHIS: if ( this->type() <= parent->type() ) { throw string("Attempt to specify a parent image that is smaller than the image itself"); }
	}
	this->parentIs(parent);
	
	/// Elements are, by default, visible.
	this->invisibleIs(false);
    }
    
    Element::~Element() {
    }
    
    string Coord::str(int prec) const {
	stringstream s;
	s << "<";
	s << setw(prec) << this->x;
	s << ",";
	s << setw(prec) << this->y;
	s << ">";
	return s.str();
    }
    
    int Element::type() const {
	// See comments in header file
	throw string("Method 'type' should be redefined by subclasses");
    }
    
    void Element::allGlyphs(std::vector<const Glyph*>& glyphs) const {
	// Obtain my sub-parts
	const std::list<Element*>& parts = this->parts();
	
	if ( parts.size() == 0 ) {
	    // This Element does not have any sub-parts, and is thus presumably
	    // a Glyph.
	    //
	    // FUTURE FIX: If Glyphs can end up having nested Glyphs
	    // (to support accents, umlauts, etc.?), then this code may
	    // not be entirely correct, because it drills all the way
	    // down, when maybe we want the "biggest" glyph?
	    const Glyph* glyph = this->asGlyph();
	    if ( glyph == NULL ) {
		// NOTE: This may not actually be an error. I can imagine a
		// situation where a Word is created, then due to subsequent
		// corrections, is reduced to 0 Glyphs. Rather than
		// complainingn, maybe we should just ignore non-glyphs?
		cerr << "ERROR: Found type " << this->type() << " when expecting a glyph" << endl;
		throw string("InternalException: Expecting a Glyph (because no sub-images exist) but Glyph not found");
	    } else {
		// Add the glyph to our list
		glyphs.push_back(glyph);
	    }
	} else {
	    // Recurse on each part.
	    for (std::list<Element*>::const_iterator it = parts.begin(); it != parts.end(); ++it) {
		(*it)->allGlyphs(glyphs);
	    }
	}
    }
    
    // FUTURE FIX: The redundancy between these two 'allGlyphs'
    // implementations should be avoided.  What to do?  Could
    // change the two signatures
    //   void Element::allGlyphs(std::vector<Glyph*> glyphs)
    //   void Element::allGlyphs(std::vector<const Glyph*> glyphs) const
    // into the single signature
    //   void Element::allGlyphs(std::vector<Glyph*> glyphs) const
    // but only by weakening separation slightly.
    
    void Element::allGlyphs(std::vector<Glyph*>& glyphs) {
	// Obtain my sub-parts
	const std::list<Element*>& parts = this->parts();
	
	if ( parts.size() == 0 ) {
	    // This Element does not have any sub-parts, and is thus presumably
	    // a Glyph.
	    //
	    // FUTURE FIX: If Glyphs can end up having nested Glyphs
	    // (to support accents, umlauts, etc.?), then this code may
	    // not be entirely correct, because it drills all the way
	    // down, when maybe we want the "biggest" glyph?
	    Glyph* glyph = this->asGlyph();
	    if ( glyph == NULL ) {
		// NOTE: This may not actually be an error. I can imagine a
		// situation where a Word is created, then due to subsequent
		// corrections, is reduced to 0 Glyphs. Rather than
		// complainingn, maybe we should just ignore non-glyphs?
		cerr << "ERROR: Found type " << this->type() << " when expecting a glyph" << endl;
		throw string("InternalException: Expecting a Glyph (because no sub-images exist) but Glyph not found");
	    } else {
		// Add the glyph to our list
		glyphs.push_back(glyph);
	    }
	} else {
	    // Recurse on each part.
	    for (std::list<Element*>::const_iterator it = parts.begin(); it != parts.end(); ++it) {
		(*it)->allGlyphs(glyphs);
	    }
	}
    }
    
    const Image* Element::pageImage() const { 
	return this->containingPage()->imageRef();
    }
    
    Page* Element::containingPage() const {
	const Element* parent = this;
	while (parent->parent() != NULL) {
	    parent = parent->parent();
	}
	const Page* result = parent->asPage();
	// FUTURE FIX: Can we avoid this const_cast without changing
	// our accessor semantics?
	return const_cast<Page*>(result);
    }
    
    void Element::registerElement(Element* image) {
	if ( image->type() <= this->type() ) {
	    // FUTURE FIX: Exception class
	    throw string("Attempt to add a larger image as a child of a smaller image");
	}
	image->parentIs(this);
	this->partsRef().push_back(image);
    }
    
    void Element::printStructure(ostream& os, const string& indent) const {
	int i = 0;
	const std::list<Element*>& parts = this->parts();
	for (std::list<Element*>::const_iterator it = parts.begin(); it != parts.end(); ++it) {
	    Element* image = *it;
	    //os << indent << setw(3) << i << ") ";
	    image->printSummary(cerr, indent, i);
	    ++i;
	}
    }
    
    void Element::printSummary(ostream& os, const string& indent, int index) const {
	// This default implementation assumes that the subclass has
	// sub-parts, and prints out one. Subclasses may redefine this
	// method, but will usually call this definition then add more
	// info.
	os << ClassName[this->type()];
	if ( index >= 0 ) { os << " #" << index; }
	os << ": ";
	os << endl;
	this->printStructure(os, indent + "  ");
    }
    
    void Element::writeText(ostream& os) const {
	// This naive implementation doesn't usually to the proper
	// thing - subclasses reimplement as necessary.
	for (ElementList::const_iterator it = this->parts().begin(); it != this->parts().end(); ++it) {
	    (*it)->writeText(os);
	}
    }

    void Element::writeImage(const std::string& filebase, const ImageArgs& config) {
	// Obtain the page image.
	const Image* image = this->pageImage();
	// Make a copy of the glyph's subregion, appropriate filterd, magnified and thresholded.
	Image im(image, this->topleft(), this->bottomright(), config);
	// Write the image to disk.
	bool verbose = false; // here preparing for future support
	im.writeGrey(filebase,verbose);
    }
    
    void Element::writeGlyphs(const string& dir, const ImageArgs& imadj) const {
	
	// Obtain the Image associated with this Element (and thus associated
	// with all Glyphs within this Element).
	const Image* image = this->pageImage();
	
	ConstGlyphLoopStart(this,glyph) {
	    // Create a new image for the sub-region of 'image' representing
	    // 'glyph'.
	    Image gim(image, glyph->topleft(), glyph->bottomright(), imadj);
	    
	    // Establish the file to write the image to.
	    //   FUTURE FIX: Create the directory 'dir' if it doesn't exist!
	    string base = dir;
	    base += "/" + glyph->id();
	    
	    // Now write gim to base (an appropriate suffix is automatically
	    // appended to 'base' by 'writeGrey')
	    gim.writeGrey(base,1);
	} GlyphLoopEnd;
    }
    
    int Element::findIndex(const Element* image) const {
	const std::list<Element*>& parts = this->parts();
	int result = -1;
	int i = 0;
	for (std::list<Element*>::const_iterator it = parts.begin(); it != parts.end(); ++it) {
	    if ( image == *it ) {
		result = i;
		break;
	    }
	    ++i;
	}
	return result;
    }
    
    string Element::id() const {
	// NOTE: This is not an efficient method - 'id' is only
	// meant to be used for debugging purposes, not core functionality.
	stringstream s;
	
	s << "id";
	const Element* image = this;
	const Element* parent = image->parent();
	while (parent != NULL) {
	    int i = parent->findIndex(image);
	    if ( i == -1 ) {
		// This should never happen - implies that the parent/parts
		// tree-structure has been corrupted.
		//DEBUG: cerr << "WARN: Failed to find an image in the parts-list of its parent" << endl;
		s << "_NA";
	    } else {
		// We have found the index of 'image' within its parent, and
		// add that index to 's'
		s << "_" << setfill('0') << setw(3) << i;
	    }
	    image = parent;
	    parent = image->parent();
	}
	return s.str();
    }
    
    char* Element::ClassName[6] = {
	"Element",
	"Page",
	"Region",
	"Line",
	"Word",
	"Glyph",
    };
    
    void Element::test(int argc, const char* argv[]) {
	cerr << "ERROR: Not unit test for Element - FIX THIS!" << endl;
    }

};
