// C++ Dependencies
using namespace std;
# include <iostream>
# include "Env.h"  
# include <string>
# include <assert.h>
# include "OCRModule.h" // OCRModule::init

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

namespace Conjecture {
    Env::Env(int argc, const char* argv[]) {
        // We record command-line flags so that Modules can process
        // their module-specific flags
        this->argcIs(argc);
        this->argvIs(argv);
        
        // Initialize "good" state to default values.
        this->verboseIs(0);
        
        // We set the default algorithm, in case command-line flags do
        // not specify one.
        this->moduleIdIs("default");
        
        // Initialize configurable parameters from command-line options.
        this->processArgs(argc, argv);
        
        // Create an instance of the appopriate OCRModule subclass -
        // this class dictates which strategy to use for every
        // algorithm, and is fundamental to the flexibility provided
        // by the Conjecture framework.
        OCRModule* module = this->makeModule();
        
        if ( module ) {
            // IMPORTANT: Must call init on 'module'. The Module
            // hierarchy cannot fully initialize itself (virtual method in
            // constructor issue).
            module->init();
        } else {
            throw string("FatalException: Failed to create an OCRModule subclass");
        }
        this->moduleIs(module);
    }
    
    Env::~Env() { }

    Page* Env::registerPage(const string& file) {
        Page* page = new Page(this, file);
        this->currentIs(page);
        this->pagesRef().push_back(page);
        return page;
    }

    bool Env::processArgs(int argc, const char** argv) {
        // This parses the 'conjecture'-level flags, NOT the
        // module-level flags.

        int i;
        const char *s1;
        bool result = false;
        /* process arguments */
        for (i = 1; i < argc; i++) {
            if (strcmp(argv[i], "--help") == 0) {
                result = false;
                goto EndArgs;
            }
            if (argv[i][0] == '-' && argv[i][1] != 0) {
                s1 = "";
                if (i + 1 < argc)
                    s1 = argv[i + 1];
                switch (argv[i][1]) {

                  case 'h': /* help */
                    result = false;
                    goto EndArgs;
                    break;

                  case 'i': 
                    // Specifies the input image file
                    this->infileIs(s1);
                    i++;
                    break;

                  case 'o': 
                    // Specifies where identified characters should be
                    // written. FUTURE FIX: We definitely do NOT want
                    // to be redirecting stdout like this. There is no
                    // need to be dumping text to (redirected) stdout
                    // when we can just open a write stream on the
                    // desired file.
                    this->outfileIs(s1);
                    i++;
                    break;

                  case 'v': /* verbose mode */
                    this->verboseIs( atoi(s1) );
                    i++;
                    break;

                  case 'V': /* verbose mode */
                    this->verboseIs( atoi(s1) );
                    i++;
                    break;

                  case 'M': 
                    // -M specifies the algorithm-set to use
                    this->moduleIdIs(s1);
                    i++;
                    break;
                    
                  case 'Z':
                    // -Z has an algset-specific semantics.  
                    // Current semantics includes:
                    //   For -A ocrad, it is a collection of command-line flags to pass
                    this->rawdataIs(s1);
                    i++;
                    break;
                    
                  default:
                    //fprintf(stderr, "# unknown option use -h for help\n");
                    i = i;
                }
                continue;
            }
            /* argument can be filename v0.2.5 */
            else if ( (argv[i][0] != '-') || ( argv[i][1] == '\0' ) ) {
                this->infileIs(argv[i]);
            }
        }
        
      EndArgs:
        if ( this->infile().size() > 0 ) {
            // If we've established an input file, we automatically
            // create a Page for it.
            this->registerPage(this->infile().c_str());
        }
        return result;
    }

    // NOTE: The Env::makeModule() is NOT defined in this class.
    // Instead, it is automatically generated by 'ocrgen' based on
    // information in the Conjecture.modules file, and placed in
    // src/custom/Modules.cc
  
    void Env::wrapup() {
        // cerr << "NOTE: Finished" << endl;
        
    }
    
    void Env::test(int argc, const char* argv[]) {
        cerr << "ERROR: Not unit test for Env - FIX THIS!" << endl;
    }

};
