using namespace std;
# include <iostream>
# include <sstream>     // stringstream
# include "ImageArgs.h"
# include "Image.h"

namespace Conjecture {
    const uint16_t ImageArgs::MaxVal = 255;
    const uint16_t ImageArgs::MinVal =   0;

    ImageArgs::ImageArgs(uint8_t magnification, uint16_t threshold, uint16_t quantize, uint8_t filter, uint8_t dustsize) {
	if ( magnification == 0 ) { magnification = 1; }
	this->magnification() = magnification;
	this->threshold()     = threshold;
	this->quantize()      = quantize;
	this->filter()        = filter;
	this->dustsize()      = dustsize;
    }

    uint16_t ImageArgs::apply(const Image* image, const Coord& pos) const {
	//DEBUG: cerr << "x=" << x << " y=" << y << " val=" << val << endl;
	
	// Obtain initial value
	int val = image->get(pos);
	
	// Perform pixel filtering
	if ( this->filter() ) {
	    // pixel.c in gocr defines the following pixel filters:
	    //   OOO  OOO  XOO  XXO  OOX  OXO
	    //   OOX  XOX  OOX  OXO  OOO  OOO
	    //   XOO  OOO  OOO  ?XX  ?XO  X?O
	    // where an O means white, X means black, and ? means either. If
	    // the 3x3 vicinity of the pixel matches any of the above, the
	    // value of the pixel is inverted.
	    //
	    // Where did these specific filters come from?  Is there research
	    // or empirical analysis to demonstrate their optimality??
	    // The 4th one is puzzling - why would we turn off the pixel?
	    //
	    // Some questions:
	    //  - the glyph boundaries are usually quite tight up against
	    //    the bounds of the character.  For example the image for
	    //    'b' will probably have no white pixels to the left.
	    //    This fact may deterimentally affact pixel filtering.
	    //    Should we always guarantee one pixel of white boundary??
	    //
	    // Here is something to experiment with.
	    //  - divide the character box into four quadrants.  design
	    //    filters that work for the top left quadrant.  For
	    //    the top right quadrant, use the same filters with
	    //    left and right columns flipped.  For the bottom left
	    //    quadrant, use the same filters as for topleft except
	    //    top and bottom rows flipped.  For bottom right rows,
	    //    rotate 180.
	    //  - this strategy allows us to design filters that will,
	    //    for example, remove pixels from the outer edges, and
	    //    add pixels on the insides.
	}
	
	// Quantize (normally this removes lower-order bits, but
	// quantize can be an arbitrary mask - val is anded with
	// the negation of quantize.
	if ( this->quantize() ) {
	    val = val & ~(this->quantize());
	}
	
	if ( this->threshold() > 0 ) {
	    // We are to convert the pixel to the extremes, white or
	    // black.  The are various things to consider here:
	    //  - transition from black to white is not linear.
	    //  - must look at the image community discussion about
	    //    gamma (typical value of 2.2) and decide whether the
	    //    input threshold should be assumed to be "subjective"
	    //    value, and properly adjusted to obtain an apparently
	    //    linear shade distribution or not.
	    //  - PAM files can have a maximum intensity anywhere from
	    //    1 to 65535. One byte is used if maxintensity is less
	    //    then 256, 2 bytes otherwise. Is the current
	    //    pam-reading code always mapping to 1 byte?
	    //  - Should we quantize the intensity values?  Could store
	    //    two pixel values per byte if we only allow 16
	    //    values. However, the more color info available, the
	    //    more potential for it to be useful (we can always
	    //    remove info if it is problematic, but cannot gain
	    //    info if it isn't there).
	    val = (val >= this->threshold()) ? MaxVal : MinVal;
	}
	return val;
    }
    
    
    string ImageArgs::str() const {
	stringstream s;
	s << "magnification=" << this->magnification()
	  << " threshold=" << this->threshold()
	  << " quantize=" << this->quantize()
	  << " filter=" << this->filter()
	  << " dustsize=" << this->dustsize();
	return s.str();
    }
};
