using namespace std;
# include <iostream>
# include <iomanip>
# include <list>

# include "StrManip.h"


// Created 2006/06/11 by wade

// Include

namespace Conjecture {

    StrManip::StrManip() {
        // Default Constructor
    }
    
    // Interface

    void StrManip::MatchStrings(const string& s, const string& t, string& s_aln, string& t_aln, char insert) {
        // This method takes two strings and determines where to insert
        // "spaces" within each so as to maximize the amount of
        // character-to-character matching between them.
        //
        // The code is based on Perl code I wrote (see the 'ocrdiff'
        // program), which in turn is based on Darwin code (see
        // http://www.biorecipes.com/DynProgBasic/code.html)
        
        // This uses a standard dynamic programming approach. In general,
        // dynamic programming has two phases - the forward phase and the
        // backward phase. In the forward phase, we compute the optimal
        // cost for each subproblem (storing results into a matrix, D). In
        // the backward phase, we reconstruct the solution that gives the
        // optimal cost.
        //
        // For our string matching problem, specifically, we will:
        //
        // (i)  use the recurrence relation to do the tabular computation of
        //      all D[i,j] (forward phase), and
        //
        // (ii) do a traceback to find the optimal alignment.
        //
        // The recurrence relation establishes a relationship between
        // D[i,j] and values of D with indices smaller than i and j. When
        // there are no smaller values of i and j then the values of D[i,j]
        // must be stated explicitly in the base conditions. The base
        // conditions are used to calculate the scores in the first row and
        // column where there are no matrix elements above and to the left.
        // This corresponds to aligning with strings of gaps. After the
        // base conditions have been established, the recurrence relation
        // is used to compute all values of D[i,j] in the table.
        //
        // The recurrence relation is:
        //
        //  D[i,j] = max{ D[i-1,j-1] + sim_mat[s[j],t[i]],
        //                D[i-1,j] + gap_score,
        //                D[i,j-1] + gap_score }
        //
        // In other words, if you have an optimal alignment up to
        // D[i-1,j-1] there are only three possibilities of what could
        // happen next:
        //
        //   1.      the characters for s[i] and t[j] match,
        //   2.      a gap is introduced in t and
        //   3.      a gap is introduced in s
        //
        // It is not possible to add a gap to both substrings. The maximum
        // of the three scores will be chosen as the optimal score and is
        // written in matrix element D[i,j].
        //
        // The second phase is to construct the optimal alignment by
        // tracing back in the matrix any path from D[m,n] to D[0,0] that
        // led to the highest score. More than one possibility can exist
        // because it is possible that there is more than one way of
        // achieving the optimal score in each matrix element D[i,j].

        int gap_score   = -1; // fix this - parameterize!
        int score       =  2; // fix this - parameterize!
        char insert_mark = '_'; // fix this - parameterize!

        unsigned n = s.size();
        unsigned m = t.size();

        // The dynamic programming matrix, with $n+1 columns of $m+1 rows.
        // String 2 forms rows, String 1 forms columns, with an extra
        // "dummy" row and column at index 0.
        int** D = new int* [m+1];  // yes, one more than size (2 more than max)
        for (unsigned i = 0; i <= m; ++i) {
            D[i] = new int [n+1];  // yes, one more than size (2 more than max)
            for (unsigned j = 0; j <= n; ++j) {
                D[i][j] = 0;
            }
        }

        //DEBUG: PrintMatrix(D,m,n);

        // A means of scoring substitutions is required. For DNA matching,
        // this is usually performed by a matrix identifying a score for
        // each of the 16 combinations of 4 DNA letters. However, for text
        // matching, using a matrix would require lots of memory (128*128
        // just for 7-bit ascii), and the matrix would be highly regular.
        // In fact, one possibility would be the identity matrix, in which
        // the score for C and C is 1, and the score for C and D (C!=D) is
        // 0. For our purposes, we do not need to use a matrix, and instead
        // use an identity scoring function.
        //
        // In addition to the matrix/scoring function, we need to assign
        // the score associated with introducing a gap. For text matching,
        // we assign a constant gap penalty of -1.
        //
        // Initialize the dynamic programming calculation using base
        // conditions. The first element of the matrix that is filled in is
        // the D[0,0] which is assigned 0. There is no penalty or score of
        // aligning nothing with nothing. Recall that to calculate matrix
        // element D[i,j], the values of D[i-1,j-1], D[i,j-1] and D[i-1,j]
        // are needed. The first row and column of the dynamic programming
        // matrix are the scores of aligning the subsequences s[0..j-1] and
        // t[0..i-1] against a string of gaps of length j or i, respectively.
        // Knowing this, the scores of the first row and column can be
        // calculated.
        //
        //   D[0,0] := 0;
        //
        // The recurrence relation for the first row and column is:
        //   D[0,j] = D[0,j-1] + gap_score;
        //   D[i,0] = D[i-1,0] + gap_score;
        //
        // Put these into loops and initialize the first row and column.

        for (unsigned j = 1; j <= n; ++j) {
            D[0][j] = D[0][j-1] + gap_score;
        }
        for (unsigned i = 1; i <= m; ++i) {
            D[i][0] = D[i-1][0] + gap_score;
        }
        
        //DEBUG: PrintMatrix(D,m,n);

        for (unsigned i = 1; i <= m; ++i) {
            for (unsigned j = 1; j <= n; ++j) {
                char c1 = s[j-1];
                char c2 = t[i-1];
                //cerr << "i=" << i << " j=" << j << " c1=" << c1 << " c2=" << c2 << endl;
                int match = D[i-1][j-1] + ((c1==c2)?score:0);
                int gaps  = D[i][j-1] + gap_score;
                int gapt = D[i-1][j] + gap_score;
                int max  = match;
                if (gaps > max) { max = gaps; }
                if (gapt > max) { max = gapt; }
                D[i][j] = max;
            }
        }
        
        // DEBUG: PrintMatrix(D,m,n);
        
        // To create the traceback, reconstruct the path from position
        // D[m+1,n+1] to D[0,0] that led to the highest score.
        
        // Start at matrix element D[m + 1,n + 1] which contains the best
        // score of the global alignment. Initialize two variable s_aln and
        // t_aln to empty strings and then concatenate the aligned
        // characters to the beginning of the string as we do the
        // traceback. The characters are concatenated to the beginning
        // because we start at the end of the strings, element D[m+1,n+1],
        // and work backwards through the strings to the starting element
        // D[0,0]. It is possible that there is not one unique path
        // backwards through the scoring matrix, i.e. some of the highest
        // scores for matching two substrings, s[j] and t[i], could have
        // come from more than one matrix element. Only one arbitrary path
        // of the set of possible paths that led to the highest score is
        // traced back through the D matrix. Although, other equivalent
        // paths may exist only one is returned.
        
        // Assign variables.
        
        unsigned i = m;
        unsigned j = n;
        
        // Keep count of the number of inserts needed in both strings
        int insert1 = 0;
        int insert2 = 0;
        
        // To traceback the path that led to the optimal score, we start at
        // the end, matrix element D[m,n] and redo the calculations in the
        // backwards direction to find out from where we came. In other
        // words, from box D[i,j] we recalculate the scores in boxes
        // D[i-1,j-1], D[i-1,j], and D[i,j-1]. By doing this we can
        // reconstruct the path from which we came. If we came from the
        // D[i-1,j-1] then we add the next character from s and t to s_aln
        // and t_aln. If we came from D[i,j-1] then we add a gap to t_aln
        // and the next character in s to s_aln. If we came from D[i-1,j],
        // then we will add a gap to s_lan and the next character in t to
        // t_aln. As this loop is written, our priority is to take the path
        // through the diagonal first (if it exists). Adding a gap to t
        // second priority and adding a gap to s third. Again, this choice
        // of priority is completely arbitrary.
        
        while ( (i > 0) && (j > 0) ) {
            char c1 = s[j-1];
            char c2 = t[i-1];
            int diff1 = D[i][j] - ((c1==c2)?score:0);
            int diff2 = D[i][j] - gap_score;
            if ( diff1 == D[i-1][j-1] ) {
                t_aln.insert(0,1,c2);
                s_aln.insert(0,1,c1);
                i = i-1;
                j = j-1;
            } else if ( diff2 == D[i][j-1] ) {
                s_aln.insert(0,1,c1);
                t_aln.insert(0,1,insert_mark);
                ++insert2;
                j = j-1;
            } else if ( diff2 == D[i-1][j] ) {
                s_aln.insert(0,1,insert_mark);
                ++insert1;
                t_aln.insert(0,1,c2);
                i = i-1;
            } else {
                throw string("should not happen");
            }
        }
        
        // At this point either i or j (or both) should be one. Finish to
        // the ends by adding gaps and the rest of the remaining string
        // until both counters, i and j, are equal to one.
        
        if (j > 0) {
            while (j > 1) {
                s_aln.insert(0,1,s[j-1]);
                t_aln.insert(0,1,insert_mark);
                ++insert2;
                j = j-1;
            }
        } else if (i > 0) {
            while (i > 0) {
                s_aln.insert(0,1,insert_mark);
                ++insert1;
                t_aln.insert(0,1,t[i-1]);
                i = i-1;
            }
        }
    }

    void StrManip::PrintMatrix(int** D, unsigned m, unsigned n) {
        for (unsigned r = 0; r <= m; ++r ) {
            int* row = D[r];
            for (unsigned c = 0; c <= n; ++c) {
                int v = row[c];
                cerr << " " << setw(3) << v;
            }
            cerr << endl;
        }
    }

    // A pig-slow hacked string-merging method. For use in debugging
    // only [wmh].
    //
    // NOTE: I am using the cumbersome 'list.insert(list.end(),val)'
    // instead of 'list.push_back(val)' because push_back generates
    // warnings under cgywin (and I refuse to allow warnings to live
    // :-) [wmh].
    string StrManip::MergeStrings(const string& str1, const string& str2, const string& indent) {
        // ****
        // Parse the strings into vectors
        list<string> v1;  // list of strings in str1
        list<string> v2;  // list of strings in str2

        unsigned pos1 = 0;  // position to start searching for next newline in str1
        unsigned pos2 = 0;  // position to start searching for next newline in str2
        unsigned max1 = 0;  // max string length in v1
        unsigned max2 = 0;  // max string length in v2
        while (1) {
            int c = 0;
            unsigned p1 = str1.find('\n',pos1);
            unsigned p2 = str2.find('\n',pos2);
            if ( p1 != string::npos ) {
                v1.insert(v1.end(),str1.substr(pos1,p1-pos1));
                pos1 = p1 + 1;
                unsigned size = v1.back().size();
                if ( size > max1 ) { max1 = size; }
            } else {
                ++c;
            }
            if ( p2 != string::npos ) {
                v2.insert(v2.end(), str2.substr(pos2,p2-pos2));
                pos2 = p2 + 1;
                unsigned size = v2.back().size();
                if ( size > max2 ) { max2 = size; }
            } else {
                ++c;
            }
            if ( c == 2 ) { break; }
        }

        // ****
        // Establish vertical spacing needed to center smaller list
        // relative to larger one.
        //  - embarassingly inefficient.
        //  - this method is for debugging only!
        int diff = v1.size() - v2.size();
        if ( diff > 0 ) {
            // v1 larger than v2 - add pad rows in v2
            unsigned top = 0; // int(diff/2);
            unsigned bot = diff; // diff - top;
            string pad(max2,' ');
            for (unsigned i = 0; i < top; ++i) { v2.insert(v2.begin(), pad); }
            for (unsigned i = 0; i < bot; ++i) { v2.insert(v2.end(), pad); }
        } else if ( diff < 0 ) {
            // v2 larger than v1 - add pad rows in v1
            diff = -diff;
            unsigned top = 0; // int(diff/2);
            unsigned bot = diff; // diff - top;
            string pad(max1,' ');
            for (unsigned i = 0; i < top; ++i) { v1.insert(v1.begin(), pad); }
            for (unsigned i = 0; i < bot; ++i) { v1.insert(v1.end(), pad);  }
        }
        if ( v1.size() != v2.size() ) {
            throw string("FatalException: Should have same-sized lists here!");
        }

        // ****
        // Merge the two lists into a single multi-line string
        string result;
        string newline("\n");
        list<string>::iterator it1 = v1.begin();
        list<string>::iterator it2 = v2.begin();
        while ( it1 != v1.end() ) {
            result += *it1 + indent + *it2 + newline;
            ++it1;
            ++it2;
        }
        return result;
    }


    // The Unit Testing Method!!!
    void StrManip::Test(int argc, const char* argv[]) {
        argc -= 2; argv = &(argv[2]);
        if ( argc > 2 ) {
            string s1(argv[1]);
            string s2(argv[2]);
            cerr << "NOTE: Comparing:" << endl
                 << "   " << s1 << endl
                 << "   " << s2 << endl;
            string res1;
            string res2;
            MatchStrings(s1,s2,res1,res2);
            cerr << "NOTE: Result is:" << endl
                 << "   " << res1 << endl
                 << "   " << res2 << endl;
        } else {
            cerr <<" NOTE: Provide two input strings to match" << endl;
        }
    }
};

