package Meta::Lib::Root;
=head1 NAME
=head1 SYNOPSIS
=head1 DESCRIPTION
=head1 NOTES
=head1 IMPLEMENTATION
=cut

###
# Never leave home without it!
use strict;

###
# Initial "class-related" static variables
use vars qw(@ISA $VERSION @EXPORT @EXPORT_OK);
$VERSION = '0.01';

###
# The superclasses (definition and importation)
@ISA = qw(Exporter);
use Exporter;

###
# The associations of this class (importation)
use IO::File;
use Carp;
use Meta::Oopl::BaseLanguage::PerlRef;

###
# The executable code of this class (methods, accessors, initializers, finalizers)


# META: NATIVE__C13__
{
  use vars qw(%PrintFormat $ElementHash $ElementCount $RightMargin);
  ## Default formatting for print_string
  %PrintFormat = (  'keyform'   => '%-20s',
                    'idxform'   => '[%3d]',
                    'keydent'   => '    ',
                    'idxdent'   => '    ',
                    'classdent' => '    ',
                    'indent'    => '    ',
                    'maxdepth'  => 8,
             );
  ## This hash maintains a list of all values printed so far.  This allows
  ##   the printing to avoid reprinting multiple copies of the identical
  ##   element.
  $ElementHash = {};
  ## This integer indicates how many elements have been printed so far.
  ##    Currently, it is flat (i.e. not dependent on how deeply nested the
  ##    printing goes), but eventually I might make the element numbering
  ##    scheme more sophisticated
  $ElementCount = 0;
  ## This integer dictates the right margin (where annotations should occur)
  $RightMargin = 80;

  #####
  ## Perl uses a hacked-up model somewhere between C++/Java and
  ## Smalltalk.  Creation of objects is separated from initialization
  ## (like Smalltalk) and one can define "pseudo class methods" for
  ## instance creation (like Smalltalk), but sending a message to a
  ## "class" object is really just invoking a method with a package name
  ## as a first argument, so there is no real inheritance of "class
  ## methods".
  ##
  ## In order to make perl more consistent with other languages while
  ## still capturing its most common instance creation mechanism, we
  ## define the pseudeo-class-method 'new' so that we can use the
  ## following object-creation syntax:
  ##     my $obj = new Person('bob');
  ## and have it look the same in Perl, C++, Java, etc.
  ##
  ## Subclasses define their own 'new' and the Meta compiler inserts
  ## calls to parent constructors automatically.
  sub new {

      my ($class) = @_;
      ## Create the class instance
      my $self = {};
      bless $self, $class;

      ## Crucial that we invoke __meta_initInstance here (it is invoked
      ## nowhere else).  Because it is virtual post-extend, and every
      ## subclass provides a definition, we are ensured that every
      ## parent component will be invoked before the class itself.
      $self->__meta_initInstance();

      ## Done!
      return $self;
  }

  sub myclass {my $res = ref($_[0]);  $res =~ s/^([^=]+)=(.*)/$1/; return $res; }

  sub __meta_extend {
      my ($self,$hash) = @_;
      ## This method is invoked within subclass '__meta_initInstance'
      ## methods to add new fields to an instance object.  Since such
      ## methods have post-extend semantics, when executing
      ## '__meta_initInstance' within a class C with parent P, the parent
      ## '__meta_initInstance' method in P is invoked before executing the
      ## '__meta_extendState' method of C, which means that the parent
      ## state is initialized before the child state.
      ##
      ## The method detects redefinitions of state, prints a warning,
      ## and overrides the field with the lower definition.
      ## FIX THIS WMH!!! Look into extending this in various ways:
      ##   - allow for multiple fields (with differing names)
      ##   - allow user control of whether a warning should be printed
      ##     when repeats found
      my $key;
      foreach $key (keys(%{$hash})) {
          printf("WARNING: Key '$key' already exists in base - overwritten in %s\n",
                 $self->myclass())
            if ( defined($self->{$key} ) );
          $self->{$key} = defined($hash->{$key}) ? $hash->{$key} : undef;
      }
  }

  sub __meta_initInstance {
      ## This method is redefined in every subclass with post-extend
      ## semantics to ensure that the parent's definition is also called.
      ## This implementation does nothing, but it is crucial that it
      ## exist so children can call it (so we don't have to worry whether
      ## the child is an immediate or distant descendent of Meta::Object)

      ## noop
  }

  sub new_copy {
    ## Default copy construct
    my ($class, $srcobj) = @_;
    my $self = { %$srcobj };
    bless $self, $class;
    return $self;
  }

  sub __meta_clone {
    ## Subclasses should implement a copy
    ## construct rather than override this method!
    my $self = $_[0];
    my $cls = ref($self);
    return $cls->new_copy($self);
  }

  sub __meta_subclassResponsibility {
      ## Given a method name and a string representing the arguments
      ## (separated by commas, with optional '=value' default value
      ## indicators and optional 'opt_' prefix (to indicated optional
      ## argument), this method generates a 'subclassResponsibility'
      ## error message so that the user knows to add an appropriately
      ## defined method in whatever subclass they are using.  The
      ## way of emulating abstract methods in Perl.
      ##
      ## EXAMPLE:
      ##   $self->subclassResponsibility('someMethod',
      ##                                 '$arg1, $arg2 = "default"');

      my ($self, $method, $args) = @_;
      printf(STDERR "ERROR: Method '$method' must be implemented in subclass '%s'.\n",
             $self->myclass());
      my $i = 1;
      printf(STDERR "   Arguments to $method are:\n       [0] \$self\n");
      foreach (split(/\s*,\s*/, $args)) {
          m/^([^=]+)(?:=\s*(.*\S))/;
          printf(STDERR "       [%d] %-20s %s\n",
                 $i, $1, (defined($2) ? ('(optional: default = \'' . $2 . '\')') : ''));
          ++$i;
      }
      exit(1);
  }

  sub splitKeywordValues {
      ## Return a hash from a string of keyword/value pairs
      my ($str, $delim) = @_;
      $delim ||= '=';

      $str = " $str";
      my $res = {};
      my @res;
      while ( $str =~ s/(.*)\s+([^\s$delim]+)\s*$delim\s*(.*\S).*/$1/ ) {
          push(@res, [$2,$3]);
      }
      foreach (reverse(@res)) {
          $res->{$_->[0]} = $_->[1];
      }
      return $res;
  }

  sub metaAsk {
      my ($question, $fp) = @_;
      $fp ||= \*STDOUT;
      my $ans;
      ## autoflush $fp 1;
      while (1) {
          print($fp "$question? ");
          $ans = <STDIN>;
          last if ( $ans =~ m/^y|n/i );
      }
      return ( $ans =~ m/^y/i );
  }

  sub askForValue {
      ## Query user interactively for input, and continue
      ## asking until the given input matches $re.  The optional
      ## third argument specifies a default value (what to use
      ## if the user presses return immediately).
      my ($question, $re, $default) = @_;
      $re ||= qr/.*/;
      $default ||= '';

      my $fullquestion = $question;
      if ( $default ) {
      $fullquestion .= " ($default)";
      }

      my $ans;
      autoflush STDOUT 1;
      while (1) {
      print(STDOUT "$fullquestion? ");
      $ans = <STDIN>;
      chomp($ans);
      $ans ||= $default if $default;
      last if ( $ans =~ m/$re/ );
      }
      return $ans;
  }


  sub metaString {
      return &print_string(@_);
  }

  sub metaPrint {
      my ($var, %format) = @_;
      my $fp = $format{fp} || \*STDOUT;
      print($fp &print_string(@_));
  }

  ####
  ## This method returns a string representing a human-readable version
  ## of the object passed in.
  sub print_string {
      my ($var, %format) = @_;
      $ElementHash = {};
      $ElementCount = 0;
      %format = &_generate_print_format(%format);
      return( &_print_string($var,'',\%format,$format{'maxdepth'}) );
  }

  ####
  ## Private method to generate print format (is it necessary to separate it?)
  sub _generate_print_format {
      my (%format) = @_;
      foreach (keys(%PrintFormat)) {
          $format{$_} = $PrintFormat{$_} unless $format{$_};
      }
      return(%format);
  }

  ####
  ## This method prints a human-readable representation of the given
  ## object, no matter what its type is.
  ##  - All lines are indented by prepending each with '$indent'
  ##    (which is usually a string of blanks).
  ##  - If the maximum recursion depth has been reached, the string
  ##    sequence '...' is printed.
  sub _print_string {
      my($var,$indent,$formref,$maxdepth) = @_;
      $indent = '' unless $indent;
      $var = 'undefined' unless defined($var);
      my $ref = ref($var);
      my $str = '';

      my $indent_length = length($indent);

      --$maxdepth;

      if ( $maxdepth == 0 ) {
          $str = "...\n";
      } elsif ( ! $ref ) {
          ++$ElementCount;
  ##       $str = &_annotateString($var,$indent_length,"E$ElementCount",$RightMargin) . "\n";
          ## Don't bother annotating simple values
          $str = "$var\n";
      } else {
          my $type;
          my($prefix,$suffix) = ("\n","");
      my $ptr = '';

          if ( $var =~ m/^(?:([^=]+)=)?([A-Z]+)\((.*)\)/ ) {
              # Class definition
              $indent = "$indent$formref->{'classdent'}";
              $str = $1 ? "CLASS $1: --> " : '';
              $ref = $2;
          $ptr = $3;
          }

          ++$ElementCount;

          if ( defined( $ElementHash->{$var} ) ) {
              $str = "   ----> E$ElementHash->{$var}\n";
              $ref = '';  ## so that it isn't processed below.
          } else {
              $ElementHash->{$var} = $ElementCount;
          }

          my $key;
          if ( $ref eq 'HASH' ) {
              $str .= &_annotateString("$ref($ptr)",$indent_length,"E$ElementCount",$RightMargin) . $prefix;

              ## NOTE: Getting _annotate to work properly for the keys will be
              ##       tricky - annotate assumes a single line, but for hashes
              ##       and arrays, we need to annotate the first line of the
              ##       return value of a recursive _print_string call.
              ##       Wmh!!! Still to be implemented.

              foreach $key (sort(keys(%$var))) {
                  $str .= sprintf("$indent$formref->{'keydent'}$formref->{'keyform'} = %s",
                                  "'$key'",
                                  &_print_string( $var->{$key},
                                                 "$indent$formref->{'keydent'}",
                                                 $formref, $maxdepth ) 
                                 );
              }
              $str .= "$suffix";
          } elsif ( $ref eq 'ARRAY' ) {
              $str .= &_annotateString("$ref($ptr)",$indent_length,"E$ElementCount",$RightMargin) . $prefix;
              foreach $key (0..$#$var) {
                  $str .= sprintf("$indent$formref->{'idxdent'}$formref->{'idxform'} = %s",
                                  $key,
                                  &_print_string( $var->[$key],
                                                 "$indent$formref->{'idxdent'}",
                                                 $formref, $maxdepth ) );
              }
              $str .= "$suffix";
          } elsif ( $ref eq 'SCALAR' ) {
              $str .= &_annotateString($ref . " $$var",$indent_length,"E$ElementCount",$RightMargin) . $prefix;
          } elsif ( $ref eq 'REF' ) {
              $str = &_print_string( "REF $$var", $indent, $formref, $maxdepth );
          } elsif ( $ref eq 'CODE' ) {
              $str = &_annotateString("FUNCTION $var",$indent_length,"E$ElementCount",$RightMargin) . "$prefix";
          } elsif ( $ref eq 'GLOB' ) {
              my $fileno = fileno($var);
              my $s = 'FileHandle ';
              if ( defined($fileno) ) {
                  $s .= "opened on fileno=$fileno";
              } else {
                  $s .= 'unopened';
              }
              $str .= &_annotateString("$s",$indent_length,"E$ElementCount",$RightMargin) . "$prefix";
          } else {
              ## $ref does not match any of the known 'built-in' types,
              ## and does not look like 'CLASS=BUILT-IN', so it must be
              ## a "special" class with its own print format, so we just
              ## print it verbatim.
              $str = &_annotateString($var,$indent_length,"E$ElementCount",$RightMargin) . "$prefix";
          }
      }
      return $str;
  }


  ##########:
  ## Purpose: Annotate a string with another string at the right margin
  ## Args   :
  ## Notes  :
  ##   This function takes a string, an indentation count, an annotation string
  ##   and a right margin index.  The string is printed, and the annotation
  ##   string is printed at the right margin.
  ##
  ## Example:
  ##    &_annotateString($var, 10, "Testing", 80);
  ## Return :

  sub _annotateString
  {
      my($result,$icnt, $annotate, $margin) = @_;

      my $len = $icnt + length($result);
      my $space = $margin - $len;
      $space = 1 unless $space > 0;

      ##$result .= sprintf("%-${space}s$annotate",' ');

      return $result;
  }    


  ## This is a hack variable to be used within System
  use vars qw($SystemResult);
}

###############################
# static METHOD 'FileLastModified'
# META: Meta.Lib.Root.FileLastModified(&#Meta!Str):@Meta!Str
sub FileLastModified {
  my $file = $_[0]; 
  #----------------------------
  # META: start of user-code
  return (stat($file))[9];
  # META: end of user-code
  #----------------------------
}

###############################
# static METHOD 'FileIsNewer'
# META: Meta.Lib.Root.FileIsNewer(&#Meta!Str,&#Meta!Str):@boolean
sub FileIsNewer {
  my $file1 = $_[0]; 
  my $file2 = $_[1]; 
  #----------------------------
  # META: start of user-code
  my $mod1 = &FileLastModified($file1);
  my $mod2 = &FileLastModified($file2);
  return ( $mod1 > $mod2 ) ? 1 : 0;
  # META: end of user-code
  #----------------------------
}

###############################
# static METHOD 'FilesDiffer'
# META: Meta.Lib.Root.FilesDiffer(&#Meta!Str,&#Meta!Str,*Meta!Map<@Meta!Str/@Meta!Str>):@boolean
sub FilesDiffer {
  my $file1 = $_[0]; 
  my $file2 = $_[1]; 
  my $options = $_[2]; if (!defined($_[2])) { $options = undef; } 
  #----------------------------
  # META: start of user-code
  $options ||= {};
  my $result;
  my $dfile = '.meta.diff';
  my $diff = Meta::Lib::Root::System("diff $file2 $file1 > $dfile", { quiet => 1, nostatus => 1} );
  if ( $? ) {
    ## Differences occured.  Now we determine if they are
    ## "real" differences.

    ## Heuristics to be applied later.
    $result = 1;
  } else {
    ## No differences.
    $result = 0;
  }
  unlink($dfile);
  return $result;
  # META: end of user-code
  #----------------------------
}

###############################
# static METHOD 'FileModTimeDiff'
# META: Meta.Lib.Root.FileModTimeDiff(&#Meta!Str,&#Meta!Str):@int
sub FileModTimeDiff {
  my $file1 = $_[0]; 
  my $file2 = $_[1]; 
  #----------------------------
  # META: start of user-code
  my $f1mt = (stat($file1))[9];
  my $f2mt = (stat($file2))[9];
  return $f2mt - $f1mt;
  # META: end of user-code
  #----------------------------
}

###############################
# static METHOD '__meta_compare'
# META: Meta.Lib.Root.__meta_compare(&#Meta!Str,&#Meta!Str,*#Meta!Map<@Meta!Str/@Meta!Str>):@int
sub __meta_compare {
  my $file = $_[0]; 
  my $validFile = $_[1]; 
  my $options = $_[2]; 
  #----------------------------
  # META: start of user-code
  my %options = %$options;

  my $verbose = $options{verbose} || 0;
  my $result;

  my $indent = $options{indent} || '';
  ##print(STDERR "NOTE: Comparing $file and $validFile ... ") if $verbose > 0;
  printf(STDERR "metac       : $indent%-25s ... ", ($options{input} || $file)) if $verbose > 0;
  ## NOTE: This system call is assuming that the shell invoked is 'sh-like'!
  ## Does `` use the users $SHELL or is it determined some other way?
  ## FIX THIS WMH!!!
  my $res = `diff $file $validFile 2>&1`;
  if ( $? ) {
    print(STDERR "failed\n") if $verbose > 0;

    if ( $options{batch} ) {
      $result = 0;
      ## No interactivity - just write the differences and return 0.
      my $diff = $validFile;
      if ( $diff !~ s/\.valid/\.diff/ ) {
        $diff .= '.diff';
      }
      my $fp = IO::File->new(">$diff") || die("ERROR: Failed to open $diff for writing - $!");
      print($fp $res);
      $fp->close();
    } else {
      ## Interaction is not disabled - we allow the user to view various
      ## diff representations, to update the validation file, etc.
      $result = 0;
      my $help = "Legal actions are:\n  d   --> display standard diff (original = <, valid = >\n  w   --> display side-by-side with 40 columns eash\n  W   --> display side-by-side with 80 columns eash\n  U   --> copy the original file into the valid file (for use when the valid file is out-of-date)\n  q   --> finish this interactive session\n  h   --> summary of options available\n  F   --> show the file\n  V   --> show the validation file\n";
      while (1) {
        print(STDERR "action> ");
        my $ans = <STDIN>;
        chomp($ans);
        if ( $ans eq 'd' ) {
          ## Show a normal diff
          system("diff $file $validFile");
        } elsif ( $ans eq 'q' ) {
          ## Quit interaction.
          last;
        } elsif ( $ans eq 'w' ) {
          ## Show a side-by-side diff, 40 columns each
          system("diff --width=80 --side-by-side --suppress-common-lines $file $validFile");
        } elsif ( $ans eq 'W' ) {
          ## Show a side-by-side diff, 80 columns each
          system("diff --width=160 --side-by-side --suppress-common-lines $file $validFile");
        } elsif ( $ans eq 'U' ) {
          ## Update the validation file to match the $file
          print(STDERR "NOTE: Updated $validFile\n");
          system("cp $file $validFile");
          $result = 2;
          last;
        } elsif ( $ans eq 'F' ) {
          system("cat $file");
        } elsif ( $ans eq 'V' ) {
          system("cat $validFile");
        } elsif ( $ans eq 'h' ) {
          print(STDERR $help);
        } else {
          print(STDERR "ERROR: Invalid input '$ans' ignored.\n");
          print(STDERR $help);
        }
      }
    }
  } else {
    ## No differences - validation succeeded.
    print(STDERR "passed\n") if ( $verbose > 0 );
    $result = 1;
  }
  return $result;
  # META: end of user-code
  #----------------------------
}

###############################
# static METHOD 'System'
# META: Meta.Lib.Root.System(&#Meta!Str,*#Meta!Map<@Meta!Str/@Meta!Str>):@int
sub System {
  my $command = $_[0]; 
  my $options = $_[1]; if (!defined($_[1])) { $options = undef; } 
  #----------------------------
  # META: start of user-code
  ## *** copied from $META_ROOT/src/lib/perl/Meta/Root.pm by Andrew D. Aug 19,04. ***
  ## %options are:
  ##    error    --> error string to use
  ##    indent   --> indent to use when printing errors
  ##    notedent --> additional spacing to put between 'NOTE' and the text.
  ##    prefix   --> prefix to use instead of 'NOTE: '
  ##    exec     --> execute (default is 1, but can be set to 0 to disable)
  ##    die      --> die on errors
  ##    quiet    --> boolean true means don't print out command
  ##    message  --> string to print for diagnostics
  ##    verify   --> boolean true means prompt for execution before executing
  ##    nostatus --> do not print out a warning if the exit status of the
  ##                 command is non-zero (by default a warning is printed, and
  ##                 execution dies if 'die' is true)

  my %options;
  %options = %$options if $options;

  chomp($command);
  my $prefix = $options{prefix} || 'NOTE:';
  my $error = $options{error} || "command failed";
  my $indent = $options{indent} || '';
  my $notedent = $options{notedent} || '';
  my $message = $options{message} || "Executing '$command'";
  my $verbose = $options{verbose} || 0;

  ## If exec is defined but is 0, we do not execute
  my $noexec = (defined($options{exec}) && (!$options{exec})) ? 1 : 0;

  ## If verify is given, we ensure that the user wants to perform this command
  if ( ( $options{verify} ) && ( ! $noexec ) ) {
    print(STDERR "${indent}${prefix} Execute '$command'? ");
    my $ans = '';
    while ( $ans !~ m/^(y|n|yes|no)$/i ) {
      $ans = <STDIN>;
      chomp($ans);
      $ans = lc($ans);
    }
    if ( $ans =~ m/^(y|yes)$/ ) {
    } else {
      $noexec = 1;
    }
    $notedent .= '   ';
  }

  printf(STDERR "${indent}${prefix}%s%s %s\n", $notedent, ($noexec ? ' (not really)' : ''), $message) unless $options{quiet};
  printf(STDERR "$(indent}${prefix}${notedent}$command\n") if $verbose >= 3;

  my $result;
  if ( $noexec ) {
    $result = 1;
  } else {
    system($command);
    $SystemResult = $?;
    if ( $SystemResult != 0 ) {
      printf(STDERR "WARNING: $error (status %d = '%s')\n     $command\n", $?, $!) unless $options{nostatus};
      die if $options{die};
    }
    $result = ! $SystemResult;
  }
  return $result;
  # META: end of user-code
  #----------------------------
}

###############################
# instance METHOD 'clone'
# META: Meta.Lib.Root.clone(#*Meta.Lib.Root):*Meta.Root
sub clone {
  my $self = $_[0]; 
  #----------------------------
  # META: start of user-code
  print(STDERR "Perl cloning not yet implemented\n");
  # META: end of user-code
  #----------------------------
}

###############################
# static METHOD '__meta_main'
# META: Meta.Lib.Root.__meta_main(@Meta!Str,*Meta!Vec<@Meta!Str>,@boolean,@int):void
sub __meta_main {
  my $exec = $_[0]; 
  my $args = $_[1]; 
  my $validate = $_[2]; 
  my $verbose = $_[3]; 

  {
    # Main method unit-testing prolog
    if ( $validate ) {
      print(STDERR "NOTE: Executing unit test for 'Root' (stdout and stderr redirected)\n");
      close(STDOUT); open(STDOUT, ">Root.out");
      close(STDERR); open(STDERR, ">Root.err");
    } 
  }
  #----------------------------
  # META: start of user-code
  my $options = {};
  &Meta::Lib::Root::__meta_compare("test", "test.valid", $options);
  # META: end of user-code
  #----------------------------
  {
    # Main method unit-testing epilog
    if ( $validate ) {
      my $tty = `tty`; chomp($tty);
      close(STDOUT); open(STDOUT, ">$tty");
      close(STDERR); open(STDERR, ">$tty");
      my $errors = &__meta_validate($validate);
      if ( ( -s 'Root.out' ) && ( $verbose > 1 ) ) { print("\nSTDOUT\n======\n"); system("cat Root.out"); } unlink('Root.out') unless ( $verbose > 2);
      if ( ( -s 'Root.err' ) && ( $verbose > 1 ) ) { print("\nSTDERR\n======\n"); system("cat Root.err"); } unlink('Root.err') unless ( $verbose > 2);
    } 
  }
}

###############################
# static METHOD '__meta_validate'
# META: Meta.Lib.Root.__meta_validate(@int):@int
sub __meta_validate {
  my $verbose = $_[0]; 
  #----------------------------
  # META: start of user-code
  my $errors = 0;
  print("NOTE:    No unit testing provided for CLASS Root\n");
  return $errors;
  # META: end of user-code
  #----------------------------
}

###############################
# static METHOD '__meta_initStatic'
# META: Meta.Lib.Root.__meta_initStatic():void
sub __meta_initStatic {
  #----------------------------
  # META: start of user-code
  # META: end of user-code
  #----------------------------
}
&__meta_initStatic();


sub __Meta_main { &__meta_main(@_); }


###
# All well-formed perl modules return true.
1;

  
