package Meta::Oopl::BaseLanguage::PerlArrayRef;
=head1 NAME
=head1 SYNOPSIS
=head1 DESCRIPTION
=head1 NOTES
=head1 IMPLEMENTATION
=cut

###
# Never leave home without it!
use strict;

###
# Initial "class-related" static variables
use vars qw(@ISA $VERSION @EXPORT @EXPORT_OK);
$VERSION = '0.01';

###
# The superclasses (definition and importation)
@ISA = qw(Meta::Oopl::BaseLanguage::PerlRef);
use Meta::Oopl::BaseLanguage::PerlRef;

###
# The associations of this class (importation)
use IO::File;
use Carp;
use Meta::Oopl::BaseLanguage::PerlRef;

###
# The executable code of this class (methods, accessors, initializers, finalizers)


# META: NATIVE__C65__
{
  sub TIEARRAY  { return bless $_[1], $_[0]; }
  sub FETCH     { return $_[0]->[$_[1]]; }
  sub STORE     { $_[0]->[$_[1]] = $_[2]; }
  sub FETCHSIZE { return scalar(@{$_[0]}) }
  sub STORESIZE { return $#{$_[0]} = $_[1]; }
  sub CLEAR     { @{$_[0]} = (); }
  sub PUSH      { push(@{shift(@_)}, @_); }
  sub POP       { return pop(@{$_[0]}); }
  sub SHIFT     { return shift(@{$_[0]}); }
  sub UNSHIFT   { return unshift(@{shift(@_)}, @_); }
  sub SPLICE    { 
    ## There appears to be what I consider a bug in 'splice':
    ##   When the second argument to 'splice' is an array, it appears
    ##   to be being evaluated in scalar context (resulting in a single
    ##   integer (the size of the array)), not expanded into an array
    ##   of multiple elements.
    ##
    ## The following code demonstrates the problem:
    ##   my @b = (2,1,30);
    ##   my @c = (2,1,30,40);
    ##   my @a = (1,2,3,4,5);
    ##   printf("Test 1: splice(\@a,1,1,20)\n");
    ##   printf("  a=(%s)\n", join(',',@a));
    ##   splice(@a, 2,1,-1);
    ##   printf("  a=(%s)\n", join(',',@a));
    ##
    ##   @a = (1,2,3,4,5);
    ##   printf("Test 2: splice(\@a,\@b) where b=(%s)\n",join(',',@b));
    ##   printf("  a=(%s)\n", join(',',@a));
    ##   splice(@a, @b);
    ##   printf("  a=(%s)\n", join(',',@a));
    ##
    ##   @a = (1,2,3,4,5);
    ##   printf("Test 3: splice(\@a,\@c) where c=(%s)\n",join(',',@c));
    ##   printf("  a=(%s)\n", join(',',@a));
    ##   splice(@a, @c);
    ##   printf("  a=(%s)\n", join(',',@a));
    ##
    ## The output of the above is:
    ##   Test 1: splice(@a,1,1,20)
    ##     a=(1,2,3,4,5)
    ##     a=(1,2,-1,4,5)
    ##   Test 2: splice(@a,@b) where b=(2,1,30)
    ##     a=(1,2,3,4,5)
    ##     a=(1,2,3)
    ##   Test 3: splice(@a,@c) where c=(2,1,30,40)
    ##     a=(1,2,3,4,5)
    ##     a=(1,2,3,4)
    ## 
    ## Because of the above, the following implementation does not work:
    ##    return splice(@{shift(@_)}, @_);
    ## 
    ## Next attempt is:
    ##    return eval "splice(\@{shift(\@_)},@_);";
    ## with the intention of interpolating @_ into a list of args
    ## within the double quotes before 'eval' sees the code.
    ## However, this does not work unless $" is ',' (and has another
    ## problem too)
    ## 
    ## Next attempt is:
    ##    $" = ','; return eval "splice(\@{shift(\@_)},@_);";
    ## which does not work because the 'shift' to remove the
    ## receiver must occur before the interpolation of @_.
    ## 
    ## Next attempt is:
    ##   $" = ','; my $self = shift; return eval "splice(\@{\$self},@_);";
    ## which finally works, but has the unintended and possibly problematic
    ## side-effect of modifying $".
    ## 
    ## Next attempt is:
    ##    my $self = shift; my $o = $"; $" = ','; my @res = eval "splice(\@{\$self},@_);"; $" = $o; return (wantarray) ? @res : $res[$#res];
    ## which works completely correctly with respect to perl, but poses two problems:
    ## 
    ##   1) Emacs fontification problems in metaoopl-mode because of the
    ##      uneven number of '"' characters (it does not pose these
    ##      problems in perl-mode or cperl-mode, so a solution is
    ##      certainly possible.
    ## 
    ##   2) Tokenization issues in the Meta parser, since ExtendPerlBlock
    ##      consumes text between matching '"' characters and this
    ##      code is not matching. Although the solution below avoids
    ##      this problem, the Lexer.pm.tmpl code has been updated to
    ##      handle $" (and related) issues. In particular,
    ##      encountering the character '$' now means "escape next
    ##      character", similar to '\'.
    ## 
    ## We could use
    ##    my $self = shift; my $o = $"; $" = ','; my @res = eval "splice(\@{\$self},@_);"; $" = $o; $o = $"; return (wantarray) ? @res : $res[$#res];
    ## to address both of the above problems (note the redundant
    ## second occurence of '$o = $"'), but I want to know that $" is
    ## being dealt with, so a slightly better approach that
    ## addresses 1) but relies on tokenization to address 2) follows.
    ## 
    ## We arrive at:
    ##    my $self = shift; my $o = $"; $" = ','; my @res = eval "splice(\@{\$self},@_);"; $" = $o; return (wantarray) ? @res : $res[$#res];
    ##    # we add a quote to fix emacs fontification "
    ## where we save $", set it to ',', obtain the result of the
    ## splice, reset $", add a spurious $" usage to balance the
    ## quotes, and return the proper context-specific result.
    ## 
    ## Although much less efficient than the preferred solution,
    ## splice is used relatively infrequently so this is probably
    ## not an issue.
    ## 
    ## FUTURE FIX: When this bug/feature in perl is addressed, we
    ## can remove all of this code and replace it with:
    ##    return splice(@{shift(@_)}, @_);
    my $self = shift; my $o = $"; $" = ','; my @res = eval "splice(\@{\$self},@_);"; $" = $o; return (wantarray) ? @res : $res[$#res];
    ## we add a quote in a comment to fix emacs fontification without affecting parsing "
  }
  sub EXTEND    { CORE::warn("ERROR: PerlArrayRef::EXTEND not yet implemented\n"); }
  ## do we have to do anything for DESTROY???  FIX THIS WMH!!!
  sub DESTROY   { }
  sub UNTIE     { &Meta::Oopl::BaseLanguage::PerlRef::report("ERROR: PerlArrayRef::UNTIE not yet implemented\n"); }
}

###############################
# static METHOD '__meta_main'
# META: Meta.Oopl.BaseLanguage.PerlArrayRef.__meta_main(@Meta!Str,*Meta!Vec<@Meta!Str>,@boolean,@int):void
sub __meta_main {
  my $exec = $_[0]; 
  my $args = $_[1]; 
  my $validate = $_[2]; 
  my $verbose = $_[3]; 

  {
    # Main method unit-testing prolog
    if ( $validate ) {
      print(STDERR "NOTE: Executing unit test for 'PerlArrayRef' (stdout and stderr redirected)\n");
      close(STDOUT); open(STDOUT, ">PerlArrayRef.out");
      close(STDERR); open(STDERR, ">PerlArrayRef.err");
    } 
  }
  #----------------------------
  # META: start of user-code
  # Main method core
  print("No RUN attribute was specified for CLASS PerlArrayRef\n");
  # META: end of user-code
  #----------------------------
  {
    # Main method unit-testing epilog
    if ( $validate ) {
      my $tty = `tty`; chomp($tty);
      close(STDOUT); open(STDOUT, ">$tty");
      close(STDERR); open(STDERR, ">$tty");
      my $errors = &__meta_validate($validate);
      if ( ( -s 'PerlArrayRef.out' ) && ( $verbose > 1 ) ) { print("\nSTDOUT\n======\n"); system("cat PerlArrayRef.out"); } unlink('PerlArrayRef.out') unless ( $verbose > 2);
      if ( ( -s 'PerlArrayRef.err' ) && ( $verbose > 1 ) ) { print("\nSTDERR\n======\n"); system("cat PerlArrayRef.err"); } unlink('PerlArrayRef.err') unless ( $verbose > 2);
    } 
  }
}

###############################
# static METHOD '__meta_validate'
# META: Meta.Oopl.BaseLanguage.PerlArrayRef.__meta_validate(@int):@int
sub __meta_validate {
  my $verbose = $_[0]; 
  #----------------------------
  # META: start of user-code
  my $errors = 0;
  print("NOTE:    No unit testing provided for CLASS PerlArrayRef\n");
  return $errors;
  # META: end of user-code
  #----------------------------
}

###############################
# instance METHOD '__meta_initInstance'
# META: Meta.Oopl.BaseLanguage.PerlArrayRef.__meta_initInstance(#*Meta.Oopl.BaseLanguage.PerlArrayRef,@int):@int
sub __meta_initInstance {
  my $self = $_[0]; 
  my $verbose = $_[1]; 
  #----------------------------
  # META: start of user-code
  $_[0]->SUPER::__meta_initInstance();
  $_[0]->__meta_extend({
  });
  # META: end of user-code
  #----------------------------
}

###############################
# static METHOD '__meta_initStatic'
# META: Meta.Oopl.BaseLanguage.PerlArrayRef.__meta_initStatic():void
sub __meta_initStatic {
  #----------------------------
  # META: start of user-code
  # META: end of user-code
  #----------------------------
}
&__meta_initStatic();


sub __Meta_main { &__meta_main(@_); }


###
# All well-formed perl modules return true.
1;

  
