package XFig::Object;
use strict;

=head1 NAME

XFig::Object -

=head1 SYNOPSIS

use XFig::Object ();

=head1 DESCRIPTION
Abstract superclass of the 7 kinds of XFig Object
1) Color Objects
2) Arc Objects
3) Compound Objects
4) Ellipse Objects
5) Polyline Objects
6) Spline Objects
7) Text Objects

=head1 NOTES


=head1 EXAMPLES


=head1 IMPLEMENTATION

Common to ARC, ELLIPSE, POLYLINE, SPLINE, TEXT
      int     sub_type                (different values for each object)
      int     pen_color               (enumeration type, pen color)
      int     depth                   (enumeration type)
      int     pen_style               (pen style, not used)


Common to ARC, ELLIPSE, POLYLINE, SPLINE
      int     line_style              (enumeration type)
      int     thickness               (1/80 inch)
      int     fill_color              (enumeration type, fill color)
      int     area_fill               (enumeration type, -1 = no fill)
      float   style_val               (1/80 inch)

Common to ARC, POLYLINE, SPLINE
      int     cap_style               (enumeration type)
      int     forward_arrow           (0: no forward arrow, 1: on)
      int     backward_arrow          (0: no forward arrow, 1: on)


Common to ARC, ELLIPSE
      int     direction               (0: clockwise, 1: counterclockwise)
      float   center_x, center_y      (center of the arc)

=cut

use Carp;                      # everyone gets Carp!
use IO::File;                  # everyone gets IO::File!
use XFig::Polyline;
use IO::File;
use XFig::Ellipse;
use XFig::Text;
use XFig::Color;
use XFig::Compound;
use XFig::Root;
use XFig::Spline;
use XFig::Arc;
use vars qw(@ISA $VERSION @EXPORT @EXPORT_OK);
@ISA = qw(XFig::Root);
$VERSION = '0.01';

####################### Package Global Variables ###################
use vars qw();




############################# Raw Code  ############################
				      
				      
########################## Class Methods ###########################

#line 1300 "fig.pxx"
sub newFrom {
    carp('Usage: XFig::Object::newFrom(class, fp, line, options)') unless (@_ >= 3);
    my($class) = $_[0]; CORE::ref($class) eq '' or carp(sprintf('XFig::Object::newFrom: ref($class) = %s instead of .', CORE::ref($class))); 
    my($fp) = $_[1]; ((!$fp) || ($fp->isa('IO::File'))) or carp(sprintf('XFig::Object::newFrom: NOT $fp->isa(IO::File)')); 
    my($line) = $_[2]; CORE::ref($line) eq '' or carp(sprintf('XFig::Object::newFrom: ref($line) = %s instead of .', CORE::ref($line))); 
    my(%options) = @_[3..$#_]; 
#line 1304 "fig.pxx"
    my $verbose = $options{verbose} || $XFig::XFig::Verbose;
    print "VERBOS:E $verbose\n";

    print(STDERR "NOTE: Creating $class instance from file...\n");

    my ($objcode, $objtype, $data);
    if ( m/^\s*(\d+)/ ) {
        $objcode = $1;
        $objtype = $XFig::XFig::Map[$objcode];
        $data    = $XFig::XFig::Objects{$objtype};
    } else {
        die("ERROR: Line $. does not represent a valid object - $_\n");
    }

    print(STDERR "NOTE: Line $. - Parsing $objtype\n") if $verbose >= 2;
    my @specs = @$data;
    my %obj;

  LINE:
    foreach my $i (0..$#specs) {
        my $spec = $specs[$i];
        my @spec = @$spec;

        if ( $#spec == -1 ) {
            # No elements in the array means that the line is a sequence of
            # <x,y> pairs.
            printf(STDERR "NOTE:    Part %d of $objtype (line %d) is collection of <x,y> points\n", $i+1, $.) if $verbose >= 3;
            my @p = split(/\s+/, $line);
            my @points;
            for (my $i = 0; $i < $#p; $i += 2) {
                push(@points, [$p[$i], $p[$i+1]]);
                printf(STDERR "NOTE:       Read <%g,%g>\n", $p[$i], $p[$i+1]) if $verbose >= 3;
            }
            $obj{points} = \@points;

        } else {
            # We use @spec to determine the keys associated with
            # each field.  @spec is an array whose even indexes
            # represent key names, and whose odd indexes represent
            # typeinfo data-structures identifying the legal
            # values the key can take on and the semantics of
            # those values.

            my ($depkey, $depval, $condkey);
            if ( $spec[0] eq 'condition' ) {
                my $cond = $spec[1];
                ($depkey, $depval, $condkey) = @$cond;
                my $currval = $obj{$depkey} || '';
                printf(STDERR "NOTE:    Part %d of $objtype (line %d) is optional based on key $depkey (%s) = %s\n", 
                       $i+1, $., $currval, $depval) if $verbose >= 3;
                shift(@spec); # Remove the conditional from this copy of specs so it isn't encountered during normal processing
                shift(@spec);
                if ( $currval eq $depval ) {
                    printf(STDERR "NOTE:       Line is valid\n") if $verbose >= 3;
                    $obj{$condkey} = {};
                } else {
                    printf(STDERR "NOTE:       Line does not exist\n") if $verbose >= 3;
                    next LINE;
                }
            } else {
                printf(STDERR "NOTE:    Part %d of $objtype (line %d) - %d simple fields\n", $i+1, $., scalar(@spec)/2) if $verbose >= 3;
            }

            # If we get here, we are to parse the current line
            my @v;
            if ( $objcode == 4 ) {
                # Special case.  Normally, we can split each line on whitespace
                # in order to identify each field.  However, for text objects,
                # the last field (the string) can contain whitespace.
                if ( $line =~ m/^((?:[0-9\.-]+\s+){13})(.*)/ ) {
                    my ($l, $s) = ($1,$2);
                    @v = (split(/\s+/, $l), $s);
                } else {
                    die("ERROR: Special parsing of string object failed at line $.");
                }
            } else {
                @v = split(/\s+/, $line);
            }

            for (my $i = 0; $i <= $#spec; $i += 2) {
                my $key = $spec[$i];
                my $typeinfo = $spec[$i+1];
                my $tiref = ref($typeinfo);
                my $v = $v[$i/2]; # remember that @v is half the size of @spec, and that $i is always even.

                # Now we make sure the value $v matches a legal value in
                # $typeinfo (and, if so, what it's semantics is).
                my $error = 0;
                my $status = &XFig::typecheckValue($v, $typeinfo, $key);
                if ( $status ) {
                    if ( $condkey ) {
                        $obj{$condkey}->{$key} = $v;
                        printf(STDERR "NOTE:       Field %-20s = %9s   %s\n", "$condkey.$key", "'$v'", $status) if $verbose >= 3;
                    } else {
                        printf(STDERR "NOTE:       Field %-20s = %9s   %s\n", $key, "'$v'", $status) if $verbose >= 3;
                        $obj{$key} = $v;
                    }
                } else {
                    die("ERROR: key $key = '$v' $status");
                }
            }
        }
        $line = &XFig::Root::nextline($fp);
    }

    ####
    # Special case code
    if ( $objcode == 6 ) {
        ###
        # The Compound object is special - the @specs array cannot
        # deal with arbitrary numbers of lines, so we hard-code
        # the processing.
        $_ = $line;
        my @objects;
        while ( ! m/^-6\s*$/ ) {
            my $obj = XFig::Object->newFrom($fp, $_, verbose => $verbose);
            push(@objects, $obj);
        }
        print(STDERR "NOTE: Encountered end of compound at line $.\n") if $verbose >= 3;
        $line = &XFig::Root::nextline($fp);
        $obj{objects} = \@objects;
    }

    $_ = $line;

    ####
    # Create the appropriate sub-instance
    my $self;
    my $sub_type = $obj{sub_type};
    my $args = \%obj;
    if ( $objcode == 0 ) {
        $self = XFig::Color->new($args);
    } elsif ( $objcode == 1 ) {
        $self = XFig::Ellipse->new($args);
    } elsif ( $objcode == 2 ) {
        $self = XFig::Polyline->new($args);
    } elsif ( $objcode == 3 ) {
        $self = XFig::Spline->new($args);
    } elsif ( $objcode == 4 ) {
        $self = XFig::Text->new($args);
    } elsif ( $objcode == 5 ) {
        $self = XFig::Arc->new($args);
    } elsif ( $objcode == 6 ) {
        $self = XFig::Compound->new($args);
    }
    return $self;
}

#line 1283 "fig.pxx"
sub new {
    carp('Usage: XFig::Object::new(class, args)') unless (@_ == 2);
    my($class) = $_[0]; CORE::ref($class) eq '' or carp(sprintf('XFig::Object::new: ref($class) = %s instead of .', CORE::ref($class))); 
    my($args) = $_[1]; ((!$args) || (CORE::ref($args) eq 'HASH')) or carp(sprintf('XFig::Object::new: ref($args) = %s instead of HASH.', ref($args))); 
#line 1284 "fig.pxx"
    my $self = $class->SUPER::new();

    print(STDERR "NOTE:    Creating $class instance from args...\n") if $XFig::XFig::Verbose > 0;

    $self->initFields($args, 'object_code');

    ###
    # We record the hash of arguments.  Sub-constructors will
    # delete from this hash as arguments are processed, so when
    # all constructors are finished, this (hidden) field will contain
    # the list of unrecognized arguments.  They will be reported
    # in 'validate'
    $self->{unknown} = $args;
    return $self;
}

########################## Instance Methods ########################

###
# Accessors
#line 1278 "fig.pxx"
sub comment              {  return ( $#_ == 0 ) ? $_[0]->{comment} : ( $_[0]->{comment} = $_[1] );  }
#line 1277 "fig.pxx"
sub incrObject_code      {  return $_[0]->{object_code} += (defined($_[1]) ? $_[1] : 1); }
#line 1277 "fig.pxx"
sub object_code          {  return ( $#_ == 0 ) ? $_[0]->{object_code} : ( $_[0]->{object_code} = $_[1] );  }
#line 1279 "fig.pxx"
sub subcomment           {  return ( $#_ == 0 ) ? $_[0]->{subcomment} : ( $_[0]->{subcomment} = $_[1] );  }
# End Accessors
###

###
# User Code
#line 1622 "fig.pxx"
sub adjustArgs {
    carp('Usage: XFig::Object::adjustArgs(args)') unless (@_ == 1);
    my($args) = $_[0]; ((!$args) || (CORE::ref($args) eq 'HASH')) or carp(sprintf('XFig::Object::adjustArgs: ref($args) = %s instead of HASH.', ref($args))); 
#line 1639 "fig.pxx"
    foreach my $t (['size', 'font_size'],
                   ['color', 'pen_color'],
                   ['bgcolor', 'fill_color'],
                   ['pattern', 'area_fill'],
                  ) {
        my $v = delete $args->{$t->[0]};
        if ( defined($v) ) {
            if ( defined( $args->{$t->[1]} ) ) {
                die(sprintf("ERROR: Keys %s and %s are ambiguous\n", @$t));
            } else {
                $args->{$t->[1]} = $v;
            }
        }
    }

    if ( ( $args->{points} ) && ( ! $args->{npoints} ) ) {
        $args->{npoints} = $#{$args->{points}}+1;
    }
}

#line 1 "GENERATED CODE"
sub initState {
    carp('Usage: XFig::Object::initState(self)') unless (@_ == 1);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::Object'))) or carp(sprintf('XFig::Object::initState: NOT $self->isa(XFig::Object)')); 
#line 3 "GENERATED CODE"
{
    $_[0]->SUPER::initState();
    $_[0]->extend({
        # subcomment => '',
        # comment => '',
        # object_code => '',
    });
}}

#line 1616 "fig.pxx"
sub translate {
    carp('Usage: XFig::Object::translate(self, dxstr, dystr)') unless (@_ == 3);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::Object'))) or carp(sprintf('XFig::Object::translate: NOT $self->isa(XFig::Object)')); 
    my($dxstr) = $_[1]; CORE::ref($dxstr) eq '' or carp(sprintf('XFig::Object::translate: ref($dxstr) = %s instead of .', CORE::ref($dxstr))); 
    my($dystr) = $_[2]; CORE::ref($dystr) eq '' or carp(sprintf('XFig::Object::translate: ref($dystr) = %s instead of .', CORE::ref($dystr))); 
#line 1617 "fig.pxx"
    my $dx = &XFig::figunits($dxstr);
    my $dy = &XFig::figunits($dystr);
    $self->rawTranslate($dx,$dy);
}

#line 1533 "fig.pxx"
sub validate {
    carp('Usage: XFig::Object::validate(self)') unless (@_ == 1);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::Object'))) or carp(sprintf('XFig::Object::validate: NOT $self->isa(XFig::Object)')); 
#line 1548 "fig.pxx"
    my $verbose = $XFig::XFig::Verbose;

    printf(STDERR "NOTE: Validating %s instance...\n", $self->myclass()) if $verbose > 0;

    ###
    # First we determine if there were any unrecognized arguments passed to
    # the constructor.
    my $unk = $self->{unknown};
    if ( %$unk ) {
        foreach my $k (keys(%$unk)) {
            my $v = $unk->{$k};
            printf(STDERR "NOTE:    Parameter %-17s = %7s was not recognized during construction\n", "'$k'", $v) if $verbose > 0;
        }
    }

    ####
    # Now obtain the object-specific field and type info and
    # ensure that all fields are valid.
    my $oc = $self->object_code();
    my $objtype = $XFig::XFig::Map[$oc];
    my $validate = $XFig::XFig::Validate{$objtype};

    # In order to detect fields that aren't specified in
    # $validate, we maintain the set of fields and delete from it
    # as we encounter valid field names.
    my %fields = %$self;
    delete $fields{unknown};

    my $errors = 0;
    my $warnings = 0;
    foreach my $fn (sort keys(%$validate)) {
        delete $fields{$fn};
        my $ti = $validate->{$fn};
        my $val = $self->{$fn};
        #print(STDERR "NOTE:    Validating '$fn' = '$val' against "); Wmh::print($ti);
        if ( ! defined($val) ) {
            print(STDERR "NOTE:    Field '$fn' is undefined\n") if $verbose > 0;
            ++$warnings;
        } else {
            my $semantics = &XFig::typecheckValue($val, $ti, $fn);
            if ( $semantics ) {
                # $val typechecked against $ti and has semantics $semantics.
                printf(STDERR "NOTE:    Field %-17s = %7s  $semantics\n", "'$fn'", "'$val'") if $verbose > 0;
            } else {
                print(STDERR "NOTE:    Field '$fn' failed typechecking\n") if $verbose > 0;
                Wmh::print($ti, \*STDERR);
                ++$errors;
            }
        }
    }

    ####
    # Check for any fields in $self that are not listed in the
    # object-specific type info.
    if ( %fields ) {
        foreach my $k (keys(%fields)) {
            my $v = $fields{$k};
            printf(STDERR "NOTE:    Field %-17s = %7s has no type information\n", "'$k'", $v) if $verbose > 0;
        }
    }

    return ( $errors ) ? 0 : 1;
}

#line 1454 "fig.pxx"
sub write {
    carp('Usage: XFig::Object::write(self, fp)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::Object'))) or carp(sprintf('XFig::Object::write: NOT $self->isa(XFig::Object)')); 
    my($fp) = $_[1]; 
#line 1455 "fig.pxx"
    my $objcode = $self->{object_code};
    my $objtype = $XFig::XFig::Map[$objcode];
    my $data    = $XFig::XFig::Objects{$objtype};

    ###
    # Each element of @specs specifies the format of a line in the
    # output of this object.
    my @specs = @$data;

    my $comment = $self->comment();
    if ( $comment ) {
        $comment =~ s/\n/\n\# /g;
        print($fp '#'x70, "\n");
        print($fp "# $comment\n");
        print($fp '#'x70, "\n");
    }
    my $subcomment = $self->subcomment();
    if ( $subcomment ) {
        $subcomment =~ s/\n/\# \n/g;
        print($fp "###\n# $subcomment\n");
    }
  LINE:
    for my $i (0..$#specs) {
        my $spec = $specs[$i];
        my @spec = @$spec;
        my @line;
        if ( $#spec == -1 ) {
            # No elements in the $spec array means that the line
            # is a sequence of <x,y> pairs, stored in the 'points'
            # field.
            push(@line, ''); # we add this dummy field to insert a space before the points (as occurs in the xfig files)
            foreach my $xy ($self->points()) {
                push(@line, @$xy);
            }
        } else {
            my ($depkey, $depval, $condkey);
            if ( $spec[0] eq 'condition' ) {
                my $cond = $spec[1];
                ($depkey,$depval,$condkey) = @$cond;
                my $currval = $self->{$depkey} || '';
                shift(@spec);
                shift(@spec);
                if ( $currval eq $depval ) {
                } else {
                    next LINE;
                }
            }

            for (my $i = 0; $i <= $#spec; $i += 2) {
                my $key = $spec[$i];
                my $typeinfo = $spec[$i+1];
                my $v;
                if ( $condkey ) {
                    $v = $self->{$condkey}->{$key};
                } else {
                    $v = $self->{$key};
                }
                if ( ! defined($v) ) {
                    print(STDERR "NOTE: Key '$key' has undef value\n");
                }
                push(@line, $v);
            }
        }
        # Each line except the first is tabbed
        print($fp (($i==0) ? '' : "\t"), join(' ', @line), "\n");
    }

    if ( $objcode == 6 ) {
        ###
        # The 'objects' array of Compound objects is not encoded in
        # the @specs arrays - we handle it specially here.
        foreach my $obj ($self->objects()) {
            $obj->write($fp);
        }
        print($fp "-6\n");
    }
}
# User Code
###

###
# Post text

# Post text
###

########################## END OF MODULE ###########################
1;

__DATA__


__END__

=head1 PUBLIC CLASS INTERFACE

=over 4

=head1 newFrom(IO::File $fp, str $line, ... %options)
###
Creates a new instances of a *subclass* of XFig::Object
based on the info in file $fp.

=cut

=head1 PUBLIC INSTANCE INTERFACE

=over 4

=head2 adjustArgs(Hash $args)
The keys expected by subclasses of XFig::Object
are predefined by XFig, but more intuitive names for
the same concepts can sometimes be given.  This
method converts Meta-introduced key names to their
corresponding XFig names.  Meta uses the following
keys related to controlling object properties:
align    --> align  (no change)
depth    --> depth  (no change)
size     --> font_size
color    --> pen_color
bgcolor  --> fill_color
pattern  --> area_fill

In addition to providing new keys, this method can
also fill in redundant keys based on other keys.
(i.e. npoints can be computed from points).

=cut

=head2 initState()
Automated method for initializing Perl class state based on field definitions.

=cut

=head2 validate()
This method is responsible for ensuring that the instance is
valid.  Subclasses that redefine should call this method
then add additional validation (i.e. this method should have
post-extend semantics).

The %Validate hash gives object-type specific hashes that map
field names to typeinfo structures (which in turn identify legal
values and the semantics of those values).

NOTE: It is assumed that the instance has been fully initialized
before this method is invoked.  This implies that 'validate' should
NOT be called during constructor invocations.

The method returns boolean true if valid, false if invalid

=cut

=head1 AUTHOR

Wade Holst

=head1 SEE ALSO

perl(1).

=cut         