package XFig::XFig;
use strict;

=head1 NAME

XFig -

=head1 SYNOPSIS

use XFig ();

=head1 DESCRIPTION
Represents an XFig file

=head1 NOTES
This class provides an object-oriented interface to the parsing
of existing .fig files and the creation of new .fig files from
perl function calls.

 $xfig = XFig::XFig->new();
 my $box = $xfig->newBox({ bounds => [0,0,500,500] });
 $box->setFillColor('blue');
 $box->comment('This is a blue box');
 my $line = $xfig->newPolyline({ points => [[0,0], [100,100], [250,100]]});
 $line->thickness(3);
 my $text = $xfig->newText( { text => 'Hello', x => 400, 'y' => 400} );
 $text->setPenColor('red');
 $xfig->write('ex.fig');

=head1 EXAMPLES


=head1 IMPLEMENTATION

=cut

use Carp;                      # everyone gets Carp!
use IO::File;                  # everyone gets IO::File!
use XFig::Polyline;
use IO::File;
use XFig::ArcBox;
use XFig::Box;
use XFig::Object;
use XFig::Polygon;
use XFig::Ellipse;
use XFig::Root;
use XFig::CenteredText;
use XFig::Text;
use vars qw(@ISA $VERSION @EXPORT @EXPORT_OK);
@ISA = qw(XFig::Root);
$VERSION = '0.01';

####################### Package Global Variables ###################
use vars qw(@Headers);

@Headers = qw(orientation justification units papersize magnification multipage transparent);
# blah
BEGIN {

    use vars qw(@Map %Types %Objects %Defaults %Validate $Debug $Verbose %PaperSizes %Colors $NewColors);

    # Applications can change this to get diagnostics
    $Verbose = 0;

    # Loaded on demand
    $NewColors = undef;

    %Defaults = (
                 angle            =>  0,    # angle of x-axis (VERIFY THIS WMH!!!)
                 area_fill        => -1,    # sometimes no fill is a good default (-1), sometimes full saturation of fill color is a good default (20).
                 arrow_height     => undef, # FIND SENSIBLE DEFAULT IF THERE IS ONE
                 arrow_style      =>  0,    # Hollow
                 arrow_thickness  =>  1,    # verify this!!!  FIX THIS WMH
                 arrow_type       =>  0,    # Stick-type (default in xfig 2.1 and earlier)
                 arrow_width      => undef, # FIND SENSIBLE DEFAULT IF THERE IS ONE
                 backward         => {},    # no backward arrow
                 backward_arrow   =>  0,    # off
                 cap_style        =>  0,    # Butt (default for xfig 2.1 and earlier)
                 depth            => 100,    # 0 is top, 999 is bottom
                 direction        =>  1,    # counterclockwise (ellipses always have this value, arcs can vary)
                 fill_color       => -1,    # Default (0 is Black)
                 font             =>  5,    # Typewriter (assuming font_flags bit 2 = 0 = latex fonts)
                 font_flags       =>  0,    # as big-endian bit sequence '0000' (non-rigid, non special, latex font, non-hidden)
                 font_size        => 12,    # as good as most, better than some
                 forward          => {},    # no forward arrow
                 forward_arrow    =>  0,    # off
                 join_style       =>  0,    # Miter (default for xfig 2.1 and earlier)
                 line_style       => -1,    # Default (0 is Solid)
                 'lrx'           => undef, # No meaningful default
                 'lry'           => undef, # No meaningful default
                 npoints          => undef, # No default makes sense.
                 pen_color        => -1,    # Default (0 is Black)
                 pen_style        =>  0,    # currently unused (always 0?)
                 objects          => undef, # No meaningful default
                 radius           => -1,    # Appears to be default for boxes at least (VERIFY THIS WMH!!!)
                 string           => undef, # No meaningful default
                 style_val        =>  1,    # verify this!!!  FIX THIS WMH
                 sub_type         => undef, # No meaningful default
                 thickness        =>  1,    # units are 1/80 inch
                 'ulx'           => undef, # No meaningful default
                 'uly'           => undef, # No meaningful default

                 # These are header defaults
                 version          => '3.2',
                 orientation      => 'Landscape',
                 justification    => 'Center',
                 units            => 'Metric',
                 papersize        => 'Letter',
                 magnification    => '100.00',
                 multipage        => 'Single',
                 transparent      => '-2',  # no transparent color
                 resolution       => 1200,
                 'coord_system'   => 2, # origin in upper_left corner
                 );
    @Map = (
                'color',      # 0
                'ellipse',    # 1
                'polyline',   # 2
                'spline',     # 3
                'text',       # 4
                'arc',        # 5
                'compound',   # 6
               );

    %PaperSizes = (
                    # Index 0 is Portrait Width (Landscape Height)
                    # Index 1 is Portrait Height (Landscape Width)

                   'Letter'  => [ '8.5in' ,  '11in' ],
                   'Legal'   => [ '8.5in' ,  '14in' ],
                   'Tabloid' => [ '17in' ,   '11in' ],
                   'A'       => [ '8.5in' ,  '11in' ],
                   'B'       => [ '11in' ,   '17in' ],
                   'C'       => [ '17in' ,   '22in' ],
                   'D'       => [ '22in' ,   '34in' ],
                   'E'       => [ '34in' ,   '44in' ],
                   'A9'      => [ '37mm' ,   '52mm' ],
                   'A8'      => [ '52mm' ,   '74mm' ],
                   'A7'      => [ '74mm' ,  '105mm' ],
                   'A6'      => [ '105mm' ,  '148mm' ],
                   'A5'      => [ '148mm' ,  '210mm' ],
                   'A4'      => [ '210mm' ,  '297mm' ],
                   'A3'      => [ '297mm' ,  '420mm' ],
                   'A2'      => [ '420mm' ,  '594mm' ],
                   'A1'      => [ '594mm' ,  '841mm' ],
                   'A0'      => [ '841mm' , '1189mm' ],
                   'B10'     => [ '32mm' ,   '45mm' ],
                   'B9'      => [ '45mm' ,   '64mm' ],
                   'B8'      => [ '64mm' ,   '91mm' ],
                   'B7'      => [ '91mm' ,  '128mm' ],
                   'B6'      => [ '128mm' ,  '182mm' ],
                   'B5'      => [ '182mm' ,  '257mm' ],
                   'B4'      => [ '257mm' ,  '364mm' ],
                   'B3'      => [ '364mm' ,  '515mm' ],
                   'B2'      => [ '515mm' ,  '728mm' ],
                   'B1'      => [ '728mm' , '1030mm' ],
                   'B0'      => [ '1030mm' , '1456mm' ],
                   );

    %Types = (
              orientation => { Landscape => 1, Portrait => 1 },
              justification => { Center  => 1, "Flush Left" => 1 },
              units         => { Metric  => 1, Inches => 1, },
              papersize     => { "Letter" => 1, "Legal" => 1, "Ledger" => 1, "Tabloid" => 1,
                                 "A" => 1, "B" => 1, "C" => 1, "D" => 1, "E" => 1,
                                 "A4" => 1,  "A3" => 1, "A2" => 1, "A1" => 1, "A0" => 1, "B5" => 1 },

              magnification => qr/^\d+\.\d+$/,
              'multipage'   => { Single => 1, Multiple => 1 },
              transparent     => qr/^(-2|-1|\d+)$/,  # -2 is None, -1 background,  0-31 standard, 32+ user-defined
              resolution      => qr/^\d+$/,
              coord_system    => qr/^\d+$/,

              color           => { -1 => 'Default',
                                   0  => 'Black',
                                   1  => 'Blue',
                                   2  => 'Green',
                                   3  => 'Cyan',
                                   4  => 'Red',
                                   5  => 'Magenta',
                                   6  => 'Yellow',
                                   7  => 'White',
                                   8  => 'Very Dark Blue',
                                   9  => 'Dark Blue',
                                   10 => 'Medium Blue',
                                   11 => 'Light Blue',
                                   12 => 'Dark Green',
                                   13 => 'Medium Green',
                                   14 => 'Light Green',
                                   15 => 'Dark Cyan',
                                   16 => 'Medium Cyan',
                                   17 => 'Light Cyan',
                                   18 => 'Dark Red',
                                   19 => 'Medium Red',
                                   20 => 'Light Red',
                                   21 => 'Dark Magenta',
                                   22 => 'Medium Magenta',
                                   23 => 'Light Magenta',
                                   24 => 'Dark Brown',
                                   25 => 'Medium Brown',
                                   26 => 'Light Brown',
                                   27 => 'Very Dark Pink',
                                   28 => 'Dark Pink',
                                   29 => 'Medium Pink',
                                   30 => 'Light Pink',
                                   31 => 'Gold',
                                   # SHOULD THESE BE UNDEF, or SOMETHING else?
                                   # The values are just provided semantics for
                                   # the keys, so it shouldn't be a problem,
                                   # but we'll have to make sure not to assume
                                   # that semantics exists.
                                   (map { $_ => undef } (32..543)),
                                 },

              user_color     => {
                                 (map { $_ => undef } (32..543)),
                                },


              display_units  => qr/^\d+(\.\d+)?$/, # 1/80 inch
              fig_units      => qr/^\d+(\.\d+)?$/, # 1200 ppi

              radians        => qr/^\d+(\.\d+)?$/,

              area_fill => {
                            -1  => 'not filled',
                            0   => 'black',
                            (map { $_ => sprintf("%d%% saturation", $_*5) } (1..19)),
                            20  => 'full saturation of color',
                            (map { $_+20 => sprintf("%d%% tinted with white", $_*5) } (1..19)),
                            40  => 'white',
                            41  => '30 degree left diagonal pattern',
                            42  => '30 degree right diagonal pattern',
                            43  => '30 degree crosshatch',
                            44  => '45 degree left diagonal pattern',
                            45  => '45 degree right diagonal pattern',
                            46  => '45 degree crosshatch',
                            47  => 'bricks',
                            48  => 'circles',
                            49  => 'horizontal lines',
                            50  => 'vertical lines',
                            51  => 'crosshatch',
                            52  => 'fish scales',
                            53  => 'small fish scales',
                            54  => 'octagons',
                            55  => 'horizontal "tire treads"',
                            56  => 'vertical "tire treads"',
                           },

              depth =>  qr/^\d{1,3}$/,  # higher numbers mean object is under objects with smaller depth

              'line_style' => {
                               -1 => 'Default',
                               0 => 'Solid',
                               1 => 'Dashed',
                               2 => 'Dotted',
                               3 => 'Dash-dotted',
                               4 => 'Dash-double-dotted',
                               5 => 'Dash-triple-dotted',
                              },

              'join_style'  => {
                                # Only for LINES
                                0 => 'Miter', # default for xfig 2.1 and earlier
                                1 => 'Bevel',
                                2 => 'Round',
                               },

              'cap_style' => {
                              # Only for LINES, ARCS and OPEN SPLINES
                              0 => 'Butt',  # default for xfig 2.1 and earlier
                              1 => 'Round',
                              2 => 'Projecting',
                             },
              'arrow_type' => {
                               # Only for LINES, ARCS, andn OPEN SPLINES
                               0 => 'Stick-type ',  # default in xfig 2.1 and earlier
                               1 => 'Closed triangle',
                               #            |\
                               #            |   \
                               #            |      \
                               #            |      /
                               #            |   /
                               #            |/
                               2 => 'Closed with "indented" butt',
                               #            |\
                               #            \   \
                               #             \     \
                               #              \       \
                               #              /       /
                               #             /     /
                               #            /   /
                               #            |/
                               3 => 'Closed with "pointed" butt',
                               #                  |\
                               #                 /    \
                               #                /        \
                               #               /            \
                               #               \            /
                               #                \        /
                               #                 \    /
                               #                  |/
                              },

              'arrow_style' => {
                                # Only for LINES, ARCS and OPEN SPLINES
                                0 => 'Hollow (actually filled with white)',
                                1 => 'Filled with pen_color',
                               },

              'hex_string'   => qr/^\#[0-9a-fA-F]+$/,

             );

    my @arrowtype = (
                     'arrow_type'      => $Types{arrow_type},
                     'arrow_style'     => $Types{arrow_style},
                     'arrow_thickness' => $Types{display_units},
                     'arrow_width'     => $Types{fig_units},
                     'arrow_height'    => $Types{fig_units},
                    );

    #####
    # This hash describes the semantics of each field of each lines
    # associated with each kind of FIG object.  Each object is described
    # by one or more lines, some whose existence is conditional on
    # the value of fields specified in previous lines.
    #
    # The hash contains object names as keys, and arrays as
    # values.  Each array contains one or more sub-arrays,
    # representing the format of successive lines.  Sub-arrays
    # provide the name of each field, followed by a
    # data-structure describing the type of the field.  If
    # there are N fields on a line, there will be 2*N elements
    # in the sub-array.  Even indices, i, represent field
    # names, and i+1 is the type info for that field.
    #
    # Fields are usually separated by whitespace in FIG files.
    # The only exception to this is that the 'string' field of
    # Text objects (the last field of the one and only line
    # associated with Text objects) may have embedded spaces.
    # This special case is handled explicitly in the
    # XFig::Object::newFrom method.
    #
    # The typeinfo associated with each field comes in three different
    # flavors:
    #
    #   1) Hash: If the typeinfo is a hash, the field's type is
    #   an enumerated type, and the keys of the hash represent
    #   the legal values of the field.  The value associated
    #   with the key provides a string describing the
    #   semantics of the value of the field.  For example, the
    #   'fill_color' field of the Polyline object has as its
    #   type, 'color', which is a hash containing the legal
    #   color values (-1 thru 543), as well as the semantics associated
    #   with each color value (i.e. color 27 is 'Very Dark Pink').
    #
    #   2) Regexp: If the typeinfo is a regexp, the field's
    #   value must match that regexp in order to be valid.
    #   Useful for ensuring a value is an integer, real,
    #   string terminated by \001, etc.
    #
    #   3) String: If the typeinfo is a string
    #   (i.e. ref($typeinfo) eq ''), the field's value must
    #   match that string exactly in order to be valid.
    %Objects = (
        'color' => [
                    # Line #1
                    [
                    'object_code'  => 0,
                    'color_number' => $Types{'user_color'},
                    'rgb'          => $Types{hex},
                    ]

                    ],


        'ellipse' => [
                      # Line #1
                      [
                      'object_code' => 1,
                      'sub_type'    => {
                          1 => 'ellipse defined by radii',
                          2 => 'ellipse defined by diameters',
                          3 => 'circle defined by radius',
                          4 => 'circle defined by diameter',
                      },
                      'line_style' => $Types{line_style},
                      'thickness'  => $Types{display_units},
                      'pen_color'  => $Types{color},
                      'fill_color' => $Types{color},
                      'depth'      => $Types{depth},
                      'pen_style'  => 0, # unused
                      'area_fill'  => $Types{area_fill},
                      'style_val'  => $Types{display_units},
                      'direction'  => 1,
                      'angle'      => $Types{radians}, # angle of x-axis
                      'center_x'   => $Types{fig_units},
                      'center_y'   => $Types{fig_units},
                      'radius_x'   => $Types{fig_units},
                      'radius_y'   => $Types{fig_units},
                      'start_x'    => $Types{fig_units},
                      'start_y'    => $Types{fig_units},
                      'end_x'      => $Types{fig_units},
                      'end_y'      => $Types{fig_units},
                      ],
                      ],

        'polyline' => [
                       # Line 1
                       [
                        'object_code'       => 2,
                        'sub_type'          => {
                            1 => 'polyline',
                            2 => 'box',
                            3 => 'polygon',
                            4 => 'arc-box',
                            5 => 'imported-picture bounding-box',
                        },
                        'line_style'     => $Types{line_style},
                        'thickness'      => $Types{display_units},
                        'pen_color'      => $Types{color},
                        'fill_color'     => $Types{color},
                        'depth'          => $Types{depth},
                        'pen_style'      => 0, # unused
                        'area_fill'      => $Types{area_fill},
                        'style_val'      => $Types{display_units},
                        'join_style'     => $Types{join_style},
                        'cap_style'      => $Types{cap_style},
                        'radius'         => qr/^(-1|$Types{display_units})$/,
                        'forward_arrow'  => {
                            0 => 'off',
                            1 => 'on',
                        },
                        'backward_arrow' => {
                            0 => 'off',
                            1 => 'on',
                        },
                        'npoints'        => qr/^\d+$/, # number of points in line
                        ],

                       # Line 2: Forward arrow line
                       [
                        condition => ['forward_arrow', 1, 'forward'],
                        @arrowtype,
                        ],

                       # Line 3: Backward arrow line
                       [
                        condition => ['backward_arrow', 1, 'backward'],
                        @arrowtype,
                        ],

                       # Line 4: Points line
                       [
                        # An empty array means the line contains an arbitrary number of x/y point pairs
                        ],

                       # Line 5: PIC line.
                       #   The FIG format documentation does not explicitly
                       #   state this, but I assume that the PIC line
                       #   only exists if 'sub_type' has value 5 (imported-picture bounding-box).
                       [
                        condition => ['sub_type', '5', 'pic'],
                        flipped   => {
                            0 => 'normal',
                            1 => 'flipped',
                        },
                        file => qr/^.*\\001$/,
                        ],

                       ],

        'spline' => [
                     # Line #1
                     [
                      'object_code'       => 3,
                      'sub_type'         => {
                          0 => 'opened approximated spline',
                          1 => 'closed approximated spline',
                          2 => 'opened interpolated spline',
                          3 => 'closed interpolated spline',
                          4 => 'opened x-spline ',
                          5 => 'closed x-spline',
                      },
                      'line_style'        => undef, # (See the end of this section)
                      'thickness'         => $Types{display_units},
                      'pen_color'         => $Types{color},
                      'fill_color'        => $Types{color},
                      'depth'             => $Types{depth},
                      'pen_style'         => 0, # unused
                      'area_fill'         => $Types{area_fill},
                      'style_val'         => $Types{display_units},
                      'cap_style'         => $Types{cap_style},
                      'forward_arrow'     => {
                          0 => 'off',
                          1 => 'on',
                      },
                      'backward_arrow'    => {
                          0 => 'off',
                          1 => 'on',
                      },
                      'npoints'           => qr/^\d+$/, # number of control points in spline
                      ],

                     # Line 2: Forward arrow line
                     [
                      condition => ['forward_arrow', 1, 'forward'],
                      @arrowtype,
                      ],

                     # Line 3: Backward arrow line
                     [
                      condition => ['backward_arrow', 1, 'backward'],
                      @arrowtype,
                      ],

                     # Line 4: Points line
                     [
                      # An empty array means the line contains an arbitrary number of x/y point pairs
                      ],

                     # Line 5: Control points line
                     [
                      # THIS IS NOT YET IMPLEMENTED!
                      hack => qr/WILL NOT MATCH/,
                      ],

                     ],

        'text' => [
                   # Line #1
                   [
                    'object_code'       => 4,
                    'sub_type'         => {
                        0 => 'Left justified',
                        1 => 'Center justified',
                        2 => 'Right justified',
                    },
                    'pen_color'        => $Types{color},  # renamed from 'color' to 'pen_color' to make more consistent with other objects
                    'depth'            => $Types{depth},
                    'pen_style'        => 0, # unused
                    'font'             => {
                           -1   => 'Default font',
                            0   => 'Times Roman (Default font in PS)',
                            1   => 'Times Italic (Roman in PS)',
                            2   => 'Times Bold (Bold in PS)',
                            3   => 'Times Bold Italic (Italic in PS)',
                            4   => 'AvantGarde Book (Sans Serif in PS)',
                            5   => 'AvantGarde Book Oblique (Typewriter in PS)',
                            6   => 'AvantGarde Demi',
                            7   => 'AvantGarde Demi Oblique',
                            8   => 'Bookman Light',
                            9   => 'Bookman Light Italic',
                            10  => 'Bookman Demi',
                            11  => 'Bookman Demi Italic',
                            12  => 'Courier',
                            13  => 'Courier Oblique',
                            14  => 'Courier Bold',
                            15  => 'Courier Bold Oblique',
                            16  => 'Helvetica',
                            17  => 'Helvetica Oblique',
                            18  => 'Helvetica Bold',
                            19  => 'Helvetica Bold Oblique',
                            20  => 'Helvetica Narrow',
                            21  => 'Helvetica Narrow Oblique',
                            22  => 'Helvetica Narrow Bold',
                            23  => 'Helvetica Narrow Bold Oblique',
                            24  => 'New Century Schoolbook Roman',
                            25  => 'New Century Schoolbook Italic',
                            26  => 'New Century Schoolbook Bold',
                            27  => 'New Century Schoolbook Bold Italic',
                            28  => 'Palatino Roman',
                            29  => 'Palatino Italic',
                            30  => 'Palatino Bold',
                            31  => 'Palatino Bold Italic',
                            32  => 'Symbol',
                            33  => 'Zapf Chancery Medium Italic',
                            34  => 'Zapf Dingbats',
                    },
                    'font_size'        => qr/^\d+$/,
                    'angle'            => $Types{radians},

                    # The font_flags field is defined as follows:
                    # Bit Description
                    # 0    Rigid text (text doesn't scale when scaling compound objects)
                    # 1    Special text (for LaTeX)
                    # 2    PostScript font (otherwise LaTeX font is used)
                    # 3    Hidden text
                    'font_flags'       => qr/^([0-9]|1[0-5])$/,
                    'height'           => $Types{fig_units},
                    'length'           => $Types{fig_units},

                    # (Fig units, coordinate of the origin
                    # of the string.  If sub_type = 0, it is
                    # the lower left corner of the string.
                    # If sub_type = 1, it is the lower
                    # center.  Otherwise it is the lower
                    # right corner of the string.)
                    'x',               => $Types{fig_units},
                    'y',               => $Types{fig_units},

                    # (ASCII characters; starts after a blank
                    # character following the last number and
                    # ends before the sequence '\001'.  This
                    # sequence is not part of the string.
                    # Characters above octal 177 are
                    # represented by \xxx where xxx is the
                    # octal value.  This permits fig files to
                    # be edited with 7-bit editors and sent
                    # by e-mail without data loss.
                    # Note that the string may contain '\n'.)
                    'string'           => qr/^.*\\001$/,
                    ],
                   ],


        'arc' => [
                  # Line #1
                  [
                   'object_code'       => 5,
                   'sub_type'       => {
                       0 => 'pie-wedge (closed)',
                       1 => 'open ended arc',
                   },
                   'line_style'     => $Types{line_style},
                   'thickness'      => $Types{display_units},  # changed name from 'line_thickness' to 'thickness' to match other objects
                   'pen_color'      => $Types{color},
                   'fill_color'     => $Types{color},
                   'depth'          => $Types{depth},
                   'pen_style'      => 0, # unused
                   'area_fill'      => $Types{area_fill},
                   'style_val'      => $Types{display_units},
                   'cap_style'      => $Types{cap_style},
                   'direction'      => {
                       0 => 'clockwise',
                       1 => 'counterclockwise',
                   },
                   'forward_arrow'  => {
                       0 => 'off',
                       1 => 'on',
                   },
                   'backward_arrow' => {
                       0 => 'off',
                       1 => 'on',
                   },
                   'center_x' => $Types{fig_units},
                   'center_y' => $Types{fig_units},
                   'x1' => $Types{fig_units},
                   'y1' => $Types{fig_units},
                   'x2' => $Types{fig_units},
                   'y2' => $Types{fig_units},
                   'x3' => $Types{fig_units},
                   'y3' => $Types{fig_units},
                   ],

                  # Line 2: Forward arrow line
                  [
                   condition => ['forward_arrow', 1, 'forward'],
                   @arrowtype,
                   ],

                  # Line 3: Backward arrow line
                  [
                   condition => ['backward_arrow', 1, 'backward'],
                   @arrowtype,
                   ],
                  ],

        'compound' => [
                       # Line 1
                       [
                        'object_code'       => 6,
                        ulx               => $Types{fig_units},
                        uly               => $Types{fig_units},
                        lrx               => $Types{fig_units},
                        lry               => $Types{fig_units},
                        ],

                       # Variable number of lines, ending with
                       # a line whose only value is '-6'.
                       # This is handled specially by the
                       # XFig::Object::newFrom method.
               ],


               );

    ######
    # This hash is a flattening of the %Objects hash.  Instead
    # of a hash of arrays of arrays (of hashes or regexps or
    # strings), we produce a hash of hashes (of hashes or
    # regexps or strings).  The primary hash maps object names to
    # hashes mapping field names to type-infos.
    #
    # After this code has been executed, $Validate{$objtype}
    # will represent a hash identifying all (most) fields of
    # objects of $objtype, along with the legal values and
    # semantics of those value.  This will be used by the
    # various 'validate' methods to ensure that instances are
    # correct.

    $Debug = 0;
    print(STDERR "NOTE: Creating Validate data-structure...\n") if $Debug;
    foreach my $k (keys(%Objects)) {
        my %data;
        my $specs = $Objects{$k};
        print(STDERR "NOTE:    Working on $k...\n") if $Debug;
        foreach my $spec (@$specs) {
            # $spec is an array of fieldname/typeinfo pairs.  Even indices, i,
            # are field-names, i+1 is typeinfo for that field name.  We convert
            # to a hash.
            my ($depkey, $depval, $condkey); # in case we have a conditional
            for (my $i = 0; $i <= $#$spec; $i += 2) {
                my $fn = $spec->[$i];
                my $ti = $spec->[$i+1];
                if ( $fn eq 'condition' ) {
                    # Special case - ignore?
                    ($depkey, $depval, $condkey) = @$ti;
                    $data{$condkey} = {};
                } elsif ( $data{$fn} ) {
                    die("ERROR: Repeat field '$fn' encountered for object '$k'");
                } elsif ( $depkey ) {
                    # All fields in this $spec line are conditional, and are
                    # to be placed in a subfield called $condkey.
                    if ( $data{$condkey}->{$fn} ) {
                        die("ERROR: Repeat field '$fn' encountered for object '$k', conditional $condkey\n");
                    } else {
                        $data{$condkey}->{$fn} = $ti;
                        print(STDERR "NOTE:       Adding $condkey->$fn (conditional on $depkey being value $depval)\n") if $Debug;
                    }
                } else {
                    print(STDERR "NOTE:       Adding $fn\n") if $Debug;
                    $data{$fn} = $ti;
                }
            }
        }
        $Validate{$k} = \%data;
    }
    $Validate{polyline}->{points} = sub { my $val = $_[0]; return ( ref($val) eq 'ARRAY' ) ? '' : 'Expecting array of point tuples'; };
    foreach my $o (qw(arc polyline spline)) {
        $Validate{$o}->{forward}  = sub { return ''; };
        $Validate{$o}->{backward} = sub { return ''; };
    }


    ####
    # We create a reverse mapping from color name to color number here.
    #   We provide numerous different versions of the color name stored
    #   in the $Types{color} hash.
    my $colorIndices = $Types{color};
    foreach my $ci (keys(%{$colorIndices})) {
        my $Color = $colorIndices->{$ci};
        next unless defined($Color); # this happens for user-defined colors
        (my $ColorNS = $Color) =~ s/\s+//g;
        my @Color = split(/\s+/, $Color);
        my $color = lc($Color);
        (my $colorNS = $color) =~ s/\s+//g;
        my $commonColor = lc($Color[0]) . join('', @Color[1..$#Color]);
        $Colors{$Color} = $ci;
        $Colors{$color} = $ci;
        $Colors{$ColorNS} = $ci;
        $Colors{$colorNS} = $ci;
        $Colors{$commonColor} = $ci;
    }
}

############################# Raw Code  ############################
				      
				      
########################## Class Methods ###########################

#line 863 "fig.pxx"
sub newFromFile {
    carp('Usage: XFig::XFig::newFromFile(class, file, options)') unless (@_ >= 2);
    my($class) = $_[0]; CORE::ref($class) eq '' or carp(sprintf('XFig::XFig::newFromFile: ref($class) = %s instead of .', CORE::ref($class))); 
    my($file) = $_[1]; CORE::ref($file) eq '' or carp(sprintf('XFig::XFig::newFromFile: ref($file) = %s instead of .', CORE::ref($file))); 
    my(%options) = @_[2..$#_]; 
#line 869 "fig.pxx"

    my $self = $class->SUPER::new();
    $self->file($file);

    my $fp = IO::File->new("<$file") || die("ERROR: Failed to open '$file' for reading.\n");
    $self->parseHeader($fp);
    $self->parseBody($fp, %options);
    $fp->close();

    return $self;
}

#line 881 "fig.pxx"
sub new {
    carp('Usage: XFig::XFig::new(class, args)') unless ( (@_ >= 1) && (@_ <= 2) );
    my($class) = $_[0]; CORE::ref($class) eq '' or carp(sprintf('XFig::XFig::new: ref($class) = %s instead of .', CORE::ref($class))); 
    my($args) = $_[1]; $args = {} unless defined($args); ((!$args) || (CORE::ref($args) eq 'HASH')) or carp(sprintf('XFig::XFig::new: ref($args) = %s instead of HASH.', ref($args))); 
#line 882 "fig.pxx"
    my $self = $class->SUPER::new();
    $self->initFields($args, 'version', @Headers, 'resolution', 'coord_system');

    return $self;
}

########################## Instance Methods ########################

###
# Accessors
#line 848 "fig.pxx"
sub addObject            {  push(@{$_[0]->{objects}}, $_[1]);  }
#line 846 "fig.pxx"
sub appendFile           {  return $_[0]->{file} .= $_[1];  }
#line 859 "fig.pxx"
sub coord_system         {  return ( $#_ == 0 ) ? $_[0]->{coord_system} : ( $_[0]->{coord_system} = $_[1] );  }
#line 846 "fig.pxx"
sub file                 {  return ( $#_ == 0 ) ? $_[0]->{file} : ( $_[0]->{file} = $_[1] );  }
#line 859 "fig.pxx"
sub incrCoord_system     {  return $_[0]->{coord_system} += (defined($_[1]) ? $_[1] : 1); }
#line 858 "fig.pxx"
sub incrResolution       {  return $_[0]->{resolution} += (defined($_[1]) ? $_[1] : 1); }
#line 857 "fig.pxx"
sub incrTransparent      {  return $_[0]->{transparent} += (defined($_[1]) ? $_[1] : 1); }
#line 850 "fig.pxx"
sub incrVersion          {  return $_[0]->{version} += (defined($_[1]) ? $_[1] : 1); }
#line 852 "fig.pxx"
sub justification        {  return ( $#_ == 0 ) ? $_[0]->{justification} : ( $_[0]->{justification} = $_[1] );  }
#line 855 "fig.pxx"
sub magnification        {  return ( $#_ == 0 ) ? $_[0]->{magnification} : ( $_[0]->{magnification} = $_[1] );  }
#line 856 "fig.pxx"
sub multipage            {  return ( $#_ == 0 ) ? $_[0]->{multipage} : ( $_[0]->{multipage} = $_[1] );  }
#line 848 "fig.pxx"
sub numObjects           {  return scalar(@{$_[0]->{objects}});  }
#line 848 "fig.pxx"
sub object               {  if ( @_ == 2 ) { return $_[0]->{objects}->[$_[1]]; } else { $_[0]->{objects}->[$_[1]] = $_[2]; return $_[2]; }  }
#line 848 "fig.pxx"
sub objects              {  if ( $#_ == 0 ) { return (wantarray) ? @{$_[0]->{objects}} : $_[0]->{objects}} else { $_[0]->{objects} = $_[1]; return (wantarray) ? @{$_[1]} : $_[1] }  }
#line 851 "fig.pxx"
sub orientation          {  return ( $#_ == 0 ) ? $_[0]->{orientation} : ( $_[0]->{orientation} = $_[1] );  }
#line 854 "fig.pxx"
sub papersize            {  return ( $#_ == 0 ) ? $_[0]->{papersize} : ( $_[0]->{papersize} = $_[1] );  }
#line 846 "fig.pxx"
sub prependFile          {  return $_[0]->{file} = $_[1] . " " . $_[0]->{file};  }
#line 858 "fig.pxx"
sub resolution           {  return ( $#_ == 0 ) ? $_[0]->{resolution} : ( $_[0]->{resolution} = $_[1] );  }
#line 857 "fig.pxx"
sub transparent          {  return ( $#_ == 0 ) ? $_[0]->{transparent} : ( $_[0]->{transparent} = $_[1] );  }
#line 853 "fig.pxx"
sub units                {  return ( $#_ == 0 ) ? $_[0]->{units} : ( $_[0]->{units} = $_[1] );  }
#line 850 "fig.pxx"
sub version              {  return ( $#_ == 0 ) ? $_[0]->{version} : ( $_[0]->{version} = $_[1] );  }
# End Accessors
###

###
# User Code
#line 1154 "fig.pxx"
sub addColor {
    carp('Usage: XFig::XFig::addColor(self, name)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::addColor: NOT $self->isa(XFig::XFig)')); 
    my($name) = $_[1]; CORE::ref($name) eq '' or carp(sprintf('XFig::XFig::addColor: ref($name) = %s instead of .', CORE::ref($name))); 
#line 1160 "fig.pxx"
    if ( ! defined($NewColors) ) {
        $NewColors = $self->parseNewColors();
        #Wmh::print($NewColors);
    }
    my $result = 0;
    my $hex = $NewColors->{$name};
    if ( $hex ) {
        # Have found the color - now we look for a free Fig color slot.
        my $slot = undef;
        foreach my $i (32..543) {
            if ( ! defined($Types{color}->{$i}) ) {
                $slot = $i;
                last;
            }
        }
        if ( defined($slot) ) {
            $Types{color}->{$slot} = $name;
            (my $tname = $name) =~ s/\s//g;
            $Colors{$name} = $slot;
            $Colors{lc($name)} = $slot;
            $Colors{$tname} = $slot;
            $Colors{lc($tname)} = $slot;

            my $colorObj = XFig::Color->new({ color_number => $slot, rgb => "#$hex" });
            my $objs = $self->objects();
            unshift(@$objs, $colorObj);

            printf("NOTE:    Adding color $name ($hex) to slot $slot\n") if $Verbose;
            $result = 1;
        } else {
            print(STDERR "ERROR: Failed to find a few color slot for $name\n");
            $result = 0;
        }
    } else {
        print(STDERR "ERROR: Failed to find color '$name' - not added\n");
        $result = 0;
    }
    return $result;
}

#line 941 "fig.pxx"
sub colorIndexFor {
    carp('Usage: XFig::XFig::colorIndexFor(color)') unless (@_ == 1);
    my($color) = $_[0]; CORE::ref($color) eq '' or carp(sprintf('XFig::XFig::colorIndexFor: ref($color) = %s instead of .', CORE::ref($color))); 
#line 955 "fig.pxx"
    my $result = ( ( $color =~ m/^\d+$/ )  && ( $Types{color}->{$color} ) ) ? $color : $XFig::XFig::Colors{$color};
    return $result;
}

#line 1135 "fig.pxx"
sub dimensions {
    carp('Usage: XFig::XFig::dimensions(self)') unless (@_ == 1);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::dimensions: NOT $self->isa(XFig::XFig)')); 
#line 1138 "fig.pxx"
    my $orient = $self->orientation();
    my $papersize = $self->papersize();
    my $dims = $PaperSizes{$papersize};
    if ( ! $dims ) {
        die("ERROR: Failed to find dimensions of papersize $papersize\n");
    } else {
        # $dims is a two-tuple containing portrait width and height.  If
        # the orientation is landscape, we need to reverse these two numbers.
        my @result = ( &figunits( $dims->[0] ), &figunits( $dims->[1] ) );
        if ( $orient eq 'Landscape' ) {
            @result = reverse(@result);
        }
        return @result;
    }
}

#line 1103 "fig.pxx"
sub figunits {
    carp('Usage: XFig::XFig::figunits(spec, resolution)') unless ( (@_ >= 1) && (@_ <= 2) );
    my($spec) = $_[0]; CORE::ref($spec) eq '' or carp(sprintf('XFig::XFig::figunits: ref($spec) = %s instead of .', CORE::ref($spec))); 
    my($resolution) = $_[1]; $resolution = 1200 unless defined($resolution); CORE::ref($resolution) eq '' or carp(sprintf('XFig::XFig::figunits: ref($resolution) = %s instead of .', CORE::ref($resolution))); $resolution =~ m/^[+-]?\d+$/ or carp(sprintf('XFig::XFig::figunits: Variable $resolution = %s does not match type int.', $resolution)); 
#line 1114 "fig.pxx"

    my $result;
    my $cmperin    = 2.54;
    if ( $spec =~ m/^(?:(\d+(?:\.\d+)?)|(\.\d+))(cm|mm|in)?$/ ) {
        my ($num, undef, $units) = ($1,$2,($3||''));
        # FIX THIS WMH!  Why are there 3 subgroups in the above regexp?
        if ( $units eq 'in' ) {
            $result = $num * $resolution;
        } elsif ( $units eq 'cm' ) {
            $result = $num * $resolution * $cmperin;
        } elsif ( $units eq 'mm' ) {
            $result = $num * $resolution * $cmperin * 10;
        } else {
            $result = $num;
        }
    } else {
        die("ERROR: Invalid unit specification '$spec'\n");
    }
    return int($result + 0.5);
}

#line 1 "GENERATED CODE"
sub initState {
    carp('Usage: XFig::XFig::initState(self)') unless (@_ == 1);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::initState: NOT $self->isa(XFig::XFig)')); 
#line 3 "GENERATED CODE"
{
    $_[0]->SUPER::initState();
    $_[0]->extend({
        # magnification => '',
        # papersize => '',
        # version => '',
        # orientation => '',
        # file => '',
        # multipage => '',
        objects => [],
        # justification => '',
        # coord_system => '',
        # resolution => '',
        # transparent => '',
        # units => '',
    });
}}

#line 925 "fig.pxx"
sub main {
    carp('Usage: XFig::XFig::main()') unless (@_ == 0);
#line 926 "fig.pxx"
    my $xfig = XFig::XFig->new();

    my $box = $xfig->newBox({ bounds => [1000,1000,5000,5000] });
    $box->setFillColor('blue');
    $box->comment('This is a blue box');

    my $line = $xfig->newPolyline({ points => [[0,0], [100,100], [250,100]] });
    $line->thickness(3);

    my $text = $xfig->newText( { text => 'Hello', x => 2000, 'y' => 500, font_size => '30'} );
    $text->setPenColor('red');

    $xfig->write('ex.fig');
}

#line 919 "fig.pxx"
sub newArc {
    carp('Usage: XFig::XFig::newArc(self, args)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::newArc: NOT $self->isa(XFig::XFig)')); 
    my($args) = $_[1]; ((!$args) || (CORE::ref($args) eq 'HASH')) or carp(sprintf('XFig::XFig::newArc: ref($args) = %s instead of HASH.', ref($args))); 
#line 920 "fig.pxx"
    my $arc = XFig::Arc->new($args);
    $self->addObject($arc);
    return $arc;
}

#line 889 "fig.pxx"
sub newBox {
    carp('Usage: XFig::XFig::newBox(self, args)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::newBox: NOT $self->isa(XFig::XFig)')); 
    my($args) = $_[1]; ((!$args) || (CORE::ref($args) eq 'HASH')) or carp(sprintf('XFig::XFig::newBox: ref($args) = %s instead of HASH.', ref($args))); 
#line 890 "fig.pxx"
    my $box = XFig::Box->new($args);
    $self->addObject($box);
    return $box;
}

#line 913 "fig.pxx"
sub newEllipse {
    carp('Usage: XFig::XFig::newEllipse(self, args)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::newEllipse: NOT $self->isa(XFig::XFig)')); 
    my($args) = $_[1]; ((!$args) || (CORE::ref($args) eq 'HASH')) or carp(sprintf('XFig::XFig::newEllipse: ref($args) = %s instead of HASH.', ref($args))); 
#line 914 "fig.pxx"
    my $ellipse = XFig::Ellipse->new($args);
    $self->addObject($ellipse);
    return $ellipse;
}

#line 907 "fig.pxx"
sub newPolygon {
    carp('Usage: XFig::XFig::newPolygon(self, args)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::newPolygon: NOT $self->isa(XFig::XFig)')); 
    my($args) = $_[1]; ((!$args) || (CORE::ref($args) eq 'HASH')) or carp(sprintf('XFig::XFig::newPolygon: ref($args) = %s instead of HASH.', ref($args))); 
#line 908 "fig.pxx"
    my $polygon = XFig::Polygon->new($args);
    $self->addObject($polygon);
    return $polygon;
}

#line 901 "fig.pxx"
sub newPolyline {
    carp('Usage: XFig::XFig::newPolyline(self, args)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::newPolyline: NOT $self->isa(XFig::XFig)')); 
    my($args) = $_[1]; ((!$args) || (CORE::ref($args) eq 'HASH')) or carp(sprintf('XFig::XFig::newPolyline: ref($args) = %s instead of HASH.', ref($args))); 
#line 902 "fig.pxx"
    my $polyline = XFig::Polyline->new($args);
    $self->addObject($polyline);
    return $polyline;
}

#line 895 "fig.pxx"
sub newText {
    carp('Usage: XFig::XFig::newText(self, args)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::newText: NOT $self->isa(XFig::XFig)')); 
    my($args) = $_[1]; ((!$args) || (CORE::ref($args) eq 'HASH')) or carp(sprintf('XFig::XFig::newText: ref($args) = %s instead of HASH.', ref($args))); 
#line 896 "fig.pxx"
    my $text = XFig::Text->new($args);
    $self->addObject($text);
    return $text;
}

#line 1063 "fig.pxx"
sub parseBody {
    carp('Usage: XFig::XFig::parseBody(self, fp, options)') unless (@_ >= 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::parseBody: NOT $self->isa(XFig::XFig)')); 
    my($fp) = $_[1]; ((!$fp) || ($fp->isa('IO::File'))) or carp(sprintf('XFig::XFig::parseBody: NOT $fp->isa(IO::File)')); 
    my(%options) = @_[2..$#_]; 
#line 1064 "fig.pxx"
    $_ = &XFig::Root::nextline($fp);
    my @body;
    my %groups;
    while (1) {
        # FIX THIS WMH!!!
        if ( m/^\s*(\d+)/ ) {
            my $objcode = $1;
            my $objtype = $Map[$objcode];
            my $objinfo = $Objects{$objtype};

            my $obj = XFig::Object->newFrom($fp, $_, %options);
            push(@body, $obj);

            my $oc = $obj->{object_code};
            $groups{$oc} ||= [];
            push(@{$groups{$oc}}, $obj);

            # NOTE: &parseObject will set $_ to represent the first line
            # of the next object.

            last unless defined($_);
        } else {
            die("ERROR: Line $. does not represent an object\n");
        }
    }
    if ( $options{verbose} ) {
        printf(STDERR "NOTE: Parsed following objects:\n") if $options{verbose};
        foreach my $oc (sort { $a <=> $b } keys(%groups)) {
            my $arr = $groups{$oc};
            printf(STDERR "NOTE:    %2d %s\n",
                   $#$arr+1, $Map[$oc]);
        }
    }
    $self->{objects} = \@body;

    return (wantarray) ? @body : \@body;

}

#line 1035 "fig.pxx"
sub parseHeader {
    carp('Usage: XFig::XFig::parseHeader(self, fp, options)') unless (@_ >= 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::parseHeader: NOT $self->isa(XFig::XFig)')); 
    my($fp) = $_[1]; ((!$fp) || ($fp->isa('IO::File'))) or carp(sprintf('XFig::XFig::parseHeader: NOT $fp->isa(IO::File)')); 
    my(%options) = @_[2..$#_]; 
#line 1038 "fig.pxx"

    my $line = <$fp>;
    if ( $line =~ m/\#FIG\s+(\d+\.\d+)/ ) {
        $self->version($1);
    }
    foreach my $header (@Headers) {
        # Obtain the value
        my $value = &XFig::Root::nextline($fp);

        # Obtain the type info and verify the value
        my $type = $Types{$header} || die("ERROR: Failed to find type info for header $header\n");
        if ( &XFig::typecheckValue($value, $type, $header) ) {
            $self->{$header} = $value;
        }
    }

    my $rcs = &XFig::Root::nextline($fp);
    if ( $rcs =~ m/^\s*(\d+)\s+(\d+)/ ) {
        $self->{resolution} = $1;
        $self->{coord_system} = $2;
    } else {
        die("ERROR: Failed to find resolution and coord_system\n");
    }
}

#line 1200 "fig.pxx"
sub parseNewColors {
    carp('Usage: XFig::XFig::parseNewColors(self)') unless (@_ == 1);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::parseNewColors: NOT $self->isa(XFig::XFig)')); 
#line 1203 "fig.pxx"
    my $file = '/usr/X11R6/lib/X11/rgb.txt';
    print(STDERR "NOTE: Loading file '$file'\n") if $Verbose;
    my $fp = IO::File->new("<$file") || die("ERROR: Failed to open '$file' for reading\n");
    my %colors;
    while (<$fp> ) {
        if ( m/^\s*(\d+)\s+(\d+)\s+(\d+)\s+(.*\S)/ ) {
            $colors{$4} = sprintf("%02x%02x%02x", $1,$2,$3);
        }
    }
    $fp->close();
    return (wantarray) ? %colors : \%colors;
}

#line 959 "fig.pxx"
sub toggleOrientation {
    carp('Usage: XFig::XFig::toggleOrientation(self)') unless (@_ == 1);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::toggleOrientation: NOT $self->isa(XFig::XFig)')); 
#line 960 "fig.pxx"
 my $o = $self->orientation(); $self->orientation( ( $o eq 'Landscape' ) ? 'Portrait' : 'Landscape' ); }

#line 992 "fig.pxx"
sub typecheckValue {
    carp('Usage: XFig::XFig::typecheckValue(value, type, field)') unless (@_ == 3);
    my($value) = $_[0]; 
    my($type) = $_[1]; 
    my($field) = $_[2]; CORE::ref($field) eq '' or carp(sprintf('XFig::XFig::typecheckValue: ref($field) = %s instead of .', CORE::ref($field))); 
#line 996 "fig.pxx"
    my $result = '';
    my $typeref = ref($type);
    if ( $typeref eq 'HASH' ) {
        # The value must match one of the keys in $type
        if ( exists $type->{$value} ) {
            $result = $type->{$value};
        } else {
            #Wmh::print($type);
            die("ERROR: Value '$value' of '$field' is not valid (does not match enumerated type values)\n");
        }
    } elsif ( $typeref eq 'Regexp' ) {
        # $value must match the regexp described by $typeref.
        if ( $value =~ m/$type/ ) {
            $result = "$type";
        } else {
            die("ERROR: Value '$value' of '$field' fails to match regexp '$type'\n");
        }
    } elsif ( $typeref eq '' ) {
        if ( $value eq $type ) {
            $result = "constant";
        } else {
            die("ERROR: Value '$value' of '$field' fails to match constant '$type'\n");
        }
    } elsif ( $typeref eq 'CODE' ) {
        # We need to evaluate the $code routine, passing $value as
        # the single argument.  The code should return '' on
        # success, a non-null string describing the error on failure.
        my $res = &$type($value);
        if ( $res ) {
            die("ERROR: Value '$value' of '$field' fails - $res");
        } else {
            $result = 'validated by code';
        }
    } else {
        die("ERROR: Unknown typeref '$typeref'\n");
    }
    return $result;
}

#line 984 "fig.pxx"
sub writeBody {
    carp('Usage: XFig::XFig::writeBody(self, fp)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::writeBody: NOT $self->isa(XFig::XFig)')); 
    my($fp) = $_[1]; 
#line 987 "fig.pxx"
    foreach my $obj ($self->objects()) {
        $obj->write($fp);
    }
}

#line 970 "fig.pxx"
sub writeHeader {
    carp('Usage: XFig::XFig::writeHeader(self, fp)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::writeHeader: NOT $self->isa(XFig::XFig)')); 
    my($fp) = $_[1]; 
#line 973 "fig.pxx"
    printf($fp "#FIG %1.1f\n", $self->version());
    foreach my $k (@Headers) {
        my $v = $self->{$k};
        if ( ! defined($v) ) {
            print(STDERR "ERROR: header '$k' is undefined!\n");
        }
        print($fp $v, "\n");
    }
    printf($fp "%s %s\n", $self->{resolution}, $self->{coord_system});
}

#line 962 "fig.pxx"
sub write {
    carp('Usage: XFig::XFig::write(self, file)') unless (@_ == 2);
    my($self) = $_[0]; ((!$self) || ($self->isa('XFig::XFig'))) or carp(sprintf('XFig::XFig::write: NOT $self->isa(XFig::XFig)')); 
    my($file) = $_[1]; CORE::ref($file) eq '' or carp(sprintf('XFig::XFig::write: ref($file) = %s instead of .', CORE::ref($file))); 
#line 963 "fig.pxx"
    my $fp = IO::File->new(">$file") || die("ERROR: Failed to open '$file' for writing\n");
    print(STDERR "NOTE: Writing $file...\n") if $Verbose;
    $self->writeHeader($fp);
    $self->writeBody($fp);
    $fp->close();
}
# User Code
###

###
# Post text

# Post text
###

########################## END OF MODULE ###########################
1;

__DATA__


__END__

=head1 PUBLIC CLASS INTERFACE

=over 4

=head1 newFromFile(str $file, ... %options)
Parses a .fig $file and creates an XFig instance
containing a number of XFig::Object subclass instances.

%options are:
verbose  --> 0 = none, 1 = minimal, 2 = some, 3 = lots

=cut

=head1 PUBLIC INSTANCE INTERFACE

=over 4

=head2 addColor(str $name)
Given the name of a color string from the 'rgb.txt' file,
this method will automatically find a free XFig color slot
and add a color object.

Returns 1 on success, 0 on failure.

=cut

=head2 colorIndexFor(str $color)
This method takes a symbolic color name and obtains the
corresponding FIG color index.  If $color is a legal color
index, it is returned verbatim.  Returns the color index on
success, returns undef on failure.

NOTE: If $color matches one of the pre-defined colors, we
can naturally return a color index.  However, if $color does
not match the current Fig colors, we could go one step
further and automatically add the color if it is recognized
from the rgb.txt file (by calling 'addColor').  However,
this would require that this be a method, not a function,
and the currently places this routine is called from do not
have access to the $xfig instance.

=cut

=head2 dimensions()
This method returns the maximum x and maximum y values for
the given paper size, orientation and resolution.

=cut

=head2 figunits(str $spec, int $resolution = 1200)
Convert various string specifications to figure units (1200
ppi).  NOTE: Currently always using 1200 ppi, even though
the 'resolution' field of XFig files may be something
different.  FIX THIS WMH!!! Could be difficult, because we
do not have the XFig instance available here, and adding a
pointer to each object to maintain it might be expensive.

$spec consists of a real number followed optionally by one
of 'cm', 'mm' or 'in.  If no suffix exists, the number is
returned as-is (converted to an integer).

=cut

=head2 initState()
Automated method for initializing Perl class state based on field definitions.

=cut

=head2 parseHeader(IO::File $fp, ... %options)
##
This method parses an XFig file header.

=cut

=head2 parseNewColors()
This method parses the rgb.txt file, providing a hash mapping
color names to hexadecimal representations.

=cut

=head2 typecheckValue(any $value, any $type, str $field)
This method returns a string indicating the sematics of
$value relative to typeinfo $typeinfo if $value is type
correct.  The method returns the empty string on failure.

=cut

=head2 writeBody(any $fp)
##
Write each object to file.

=cut

=head2 writeHeader(any $fp)
##
Write the header to file

=cut

=head1 AUTHOR

Wade Holst

=head1 SEE ALSO

perl(1).

=cut         
